import argparse
from distributed import Client,LocalCluster
import multiprocessing
import pickle

from pysmFISH import dots_coords_calculations
from pysmFISH import utils



def dots_coords_correction():
    """
    This script is used to collect all the raw countings from the different
    hybridization, correct the coords according to the registration of the 
    reference gene and remove the dots that overlap in the overlapping
    regions between the images. Save the aggregate coords and also the coords after dots processing

    Input via argparse

    Parameters:
    -----------

    path: string. 
        Exact path to the experiment folder
    pxl: int 
        Radius of pixel used to create the neighbourhood (nhood) used to define 
        when two dots are the same
    
    """

    # Inputs of the function
    parser = argparse.ArgumentParser(description='Dots coords consolidation \
                                    and correction')

    parser.add_argument('-path', help='path to the experiment folder')
    parser.add_argument('-pixel_radius', help='adius of pixel used to create the nhood \
                            that is used to define that two pixels are the same', 
                            type=int)
    parser.add_argument('-scheduler', default=False, help='dask scheduler address ex. tcp://192.168.0.4:7003')
    
    args = parser.parse_args()

    # retrieve the parameters
    processing_experiment_directory = args.path
    pxl = args.pixel_radius

    # Dask scheduler address
    scheduler_address = args.scheduler

    if scheduler_address:
        # Start dask client on server or cluster
        client=Client(scheduler_address)

    else:
        # Start dask client on local machine. It will use all the availabe
        # cores -1

        # number of core to use
        ncores = multiprocessing.cpu_count()-1
        cluster = LocalCluster(n_workers=ncores)
        client=Client(cluster)

    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the experiment directory
    processing_experiment_directory=utils.check_trailing_slash(processing_experiment_directory,os_windows)

    stitched_reference_files_dir = processing_experiment_directory + 'stitched_reference_files'

    # Check training slash in the stitched reference directory
    stitched_reference_files_dir=utils.check_trailing_slash(stitched_reference_files_dir,os_windows)

    # Collect the infos of the experiment and the processing
    # Parse the Experimental metadata file (serial)
    experiment_infos,image_properties, hybridizations_infos, \
    converted_positions, microscope_parameters =\
    utils.experimental_metadata_parser(processing_experiment_directory)

    # Parse the configuration file 
    flt_rawcnt_config = utils.filtering_raw_counting_config_parser(processing_experiment_directory)


    # get the reference gene
    reference_gene = flt_rawcnt_config['reference_gene']

    # get the overlapping percentage and image_size
    overlapping_percentage = image_properties['Overlapping_percentage']

    # Consider a square image
    image_size = image_properties['HybImageSize']['columns']

    # Combine all counts
    all_raw_counts = dots_coords_calculations.combine_raw_counting_results(flt_rawcnt_config,
                                    hybridizations_infos,experiment_infos,
                                    processing_experiment_directory,stitched_reference_files_dir,
                                    reference_gene,add_slash)

    # Create a dictionary with only the selected peaks coords after alignment
    aligned_peaks_dict = all_raw_counts['selected_peaks_coords_aligned']

    # Create list of tuples to process each hybridization/gene on a different core
    combinations = dots_coords_calculations.processing_combinations(list(hybridizations_infos.keys()),aligned_peaks_dict)

    # Add corresponding registration_data and the corresponding coords files to the
    # tuple is order to recduce the size of the info transferred in the newtwork
    added_combinations =list()
    for idx,combination in enumerate(combinations):
        hybridization = combination[0]
        gene = combination[1]
        reg_data_combination = all_raw_counts['registration_data'][hybridization]
        aligned_peaks_dict_gene = all_raw_counts['selected_peaks_coords_aligned'][hybridization][gene]
        combination_dict = {
                'hybridization':hybridization,
                'gene':gene,
                'reg_data_combination':reg_data_combination,
                'aligned_peaks_dict_gene': aligned_peaks_dict_gene
        }
        added_combinations.append(combination_dict)

        # Process each gene in parallel
        futures_processes = client.map(dots_coords_calculations.function_to_run_dots_removal_parallel,added_combinations,
                            overlapping_percentage = overlapping_percentage,
                            image_size = image_size,pxl = pxl)

        cleaned_dots_list = client.gather(futures_processes)

    # Convert the list of dictionaries in one single dictionary
    # The saved dictionary cotains all the dots, the reference to the tile pos
    # has been removed during the overlapping dots removal step

    all_countings = dict()
    all_countings['all_coords_cleaned'] = dict()
    all_countings['all_coords'] = dict()
    all_countings['removed_coords'] = dict()

    for el in cleaned_dots_list:
        hybridization = list(el.keys())[0]
        gene = list(el[hybridization].keys())[0]
        
        renamed_gene = gene + '_' + hybridization
        
        all_countings['all_coords_cleaned'][renamed_gene] = el[hybridization][gene]['all_coords_cleaned']
        all_countings['all_coords'][renamed_gene] = el[hybridization][gene]['all_coords']
        all_countings['removed_coords'][renamed_gene] = el[hybridization][gene]['removed_coords']

    # Save all the data
    counting_data_name = processing_experiment_directory +experiment_infos['ExperimentName']+'_all_cleaned_raw_counting_data.pkl'
    pickle.dump(all_countings,open(counting_data_name,'wb'))

    client.close()


if __name__ == "__main__":
    dots_coords_correction()


# Import dask distributed to monitor the running client
from distributed import Client,LocalCluster
from dask import delayed
from collections import OrderedDict
import os
import logging
import glob
import nd2reader as nd2
from skimage import io, img_as_float,filters
import pickle
import argparse
import h5py
import yaml
import sys
import multiprocessing
import shutil
# -----------------
import numpy as  np

from pysmFISH import utils,counting,dots_calling,io, hdf5_utils

from pysmFISH.stitching_package import stitching
from pysmFISH.stitching_package import tilejoining
from pysmFISH.stitching_package import hdf5preparation
from pysmFISH.stitching_package import pairwisesingle


def preprocessing_script():
    """
    This script will process all the hybridization folders combined in a 
    processing folder. The input parameters are passed using arparse

    Parameters:
    -----------
    
    scheduler: string
        tcp address of the dask.distributed scheduler (ex. tcp://192.168.0.4:7003). 
        default = False. If False the process will run on the local computer using nCPUs-1

    path: string
        Path to the processing directory


    """


    # Inputs of the function
    parser = argparse.ArgumentParser(description='Preprocessing script')
    parser.add_argument('-scheduler', default=False, help='dask scheduler address ex. tcp://192.168.0.4:7003')
    parser.add_argument('-path', help='processing directory')
    args = parser.parse_args()
    
    # Directory to process
    processing_directory = args.path
    # Dask scheduler address
    scheduler_address = args.scheduler
    
    if scheduler_address:
        # Start dask client on server or cluster
        client=Client(scheduler_address)

    else:
        # Start dask client on local machine. It will use all the availabe
        # cores -1

        # number of core to use
        ncores = multiprocessing.cpu_count()-1
        cluster = LocalCluster(n_workers=ncores)
        client=Client(cluster)

    # Subdirectories of the processing_directory that need to be skipped for the
    # analysis
    blocked_directories = ['_logs']

    # Starting logger
    utils.init_file_logger(processing_directory)
    logger = logging.getLogger()

    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the processing directory
    processing_directory=utils.check_trailing_slash(processing_directory,os_windows)

    # Get a list of the hybridization to process
    processing_hyb_list = next(os.walk(processing_directory))[1]

    # Remove the blocked directories from the directories to process
    processing_hyb_list = [el for el in processing_hyb_list if el not in blocked_directories ]

    for processing_hyb in processing_hyb_list:
    
        # Determine the hyb number from the name
        hybridization_number = processing_hyb.split('_hyb')[-1]
        hybridization = 'Hybridization' + hybridization_number
        hyb_dir = processing_directory + processing_hyb + add_slash
        
        # Parse the Experimental metadata file (serial)
        experiment_infos,image_properties, hybridizations_infos, \
        converted_positions, microscope_parameters =\
        utils.experimental_metadata_parser(hyb_dir)
        
        # Parse the configuration file 
        flt_rawcnt_config = utils.filtering_raw_counting_config_parser(hyb_dir)
        
        
        # ----------------- .nd2 FILE CONVERSION ------------------------------

        # Create the temporary subdirectory tree (serial)
        tmp_dir_path, tmp_gene_dirs=utils.create_subdirectory_tree(hyb_dir,\
                    hybridization,hybridizations_infos,processing_hyb,suffix='tmp',add_slash=add_slash)

        # Get the list of the nd2 files to process inside the directory
        files_list = glob.glob(hyb_dir+processing_hyb+'_raw_data'+add_slash+'*.nd2')

        # Get the list of genes that are analyzed in the current hybridization
        gene_list = list(hybridizations_infos[hybridization].keys())

        # Organize the file to process in a list which order match the gene_list for
        # parallel processing
        organized_files_list = [f for gene in gene_list for f in files_list if gene+'.nd2' in f  ]
        organized_tmp_dir_list = [f for gene in gene_list for f in tmp_gene_dirs if gene in f  ]

        # Each .nd2 file will be processed in a worker part of a different node
        # Get the addresses of one process/node to use for conversion
        node_addresses = utils.identify_nodes(client)
        workers_conversion = [list(el.items())[0][1] for key,el in node_addresses.items()]

        # Run the conversion
        futures_processes=client.map(io.nd2_to_npy,gene_list,organized_files_list,
                                    tmp_gene_dirs,processing_hyb=processing_hyb,
                                    use_ram=flt_rawcnt_config['use_ram'],
                                    max_ram=flt_rawcnt_config['max_ram'],
                                    workers=workers_conversion)
        client.gather(futures_processes)

        

        # ---------------------------------------------------------------------
        
        
        # ----------------- FILTERING AND RAW COUNTING ------------------------
        
        # Create directories 

        # Create the directory where to save the filtered images
        suffix = 'filtered_png'
        filtered_png_img_dir_path, filtered_png_img_gene_dirs = \
                utils.create_subdirectory_tree(hyb_dir,hybridization,hybridizations_infos,
                            processing_hyb,suffix,add_slash,analysis_name=flt_rawcnt_config['analysis_name'])

        suffix = 'filtered_npy'
        filtered_img_dir_path, filtered_img_gene_dirs = \
                utils.create_subdirectory_tree(hyb_dir,hybridization,hybridizations_infos,
                            processing_hyb,suffix,add_slash,analysis_name=flt_rawcnt_config['analysis_name'])

        # Create the directory where to save the counting
        suffix = 'counting'
        counting_dir_path, counting_gene_dirs = \
            utils.create_subdirectory_tree(hyb_dir,hybridization,hybridizations_infos,processing_hyb,
                            suffix,add_slash,flt_rawcnt_config['skip_tags_counting'],
                            flt_rawcnt_config['skip_genes_counting'],
                            analysis_name=flt_rawcnt_config['analysis_name'])


        if flt_rawcnt_config['illumination_correction']:

            # Create the directory where to save the counting
            suffix = 'illumination_funcs'
            illumination_func_dir_path, illumination_func_gene_dirs = \
                utils.create_subdirectory_tree(hyb_dir,hybridization,hybridizations_infos,processing_hyb,
                                                suffix,add_slash,analysis_name=flt_rawcnt_config['analysis_name'])

            # Loop through channels and calculate illumination
            for gene in hybridizations_infos[hybridization].keys():
                
                flist_img_to_filter=glob.glob(hyb_dir+processing_hyb+'_tmp/'+processing_hyb+'_'+gene+'_tmp/*.npy')

                logger.debug('Create average image for gene %s', gene)

                # Chunking the image list
                num_chunks = sum(list(client.ncores().values()))
                chunked_list = utils.list_chunking(flist_img_to_filter,num_chunks)

                # Scatter the images sublists to process in parallel
                futures = client.scatter(chunked_list)

                # Create dask processing graph
                output = []
                for future in futures:
                    ImgMean = delayed(utils.partial_image_mean)(future)
                    output.append(ImgMean)
                ImgMean_all = delayed(sum)(output)
                ImgMean_all = ImgMean_all/float(len(futures))

                # Compute the graph
                ImgMean = ImgMean_all.compute()

                logger.debug('Create illumination function for gene %s',gene)
                # Create illumination function
                Illumination=filters.gaussian(ImgMean,sigma=(20,300,300))

                # Normalization of the illumination
                Illumination_flat=np.amax(Illumination,axis=0)
                Illumination_norm=Illumination_flat/np.amax(Illumination_flat)

                logger.debug('Save illumination function for gene %s',gene)
                # Save the illumination function
                illumination_path = [ill_path for ill_path in illumination_func_gene_dirs if gene in ill_path][0]
                illumination_fname=illumination_path+gene+'_illumination_func.npy'
                np.save(illumination_fname,Illumination_norm,allow_pickle=False)  

                # Broadcast the illumination function to all the cores
                client.scatter(Illumination_norm, broadcast=True)

                logger.debug('Filtering %s',gene)
                # Filtering and counting
                futures_processes=client.map(counting.filtering_and_counting_ill_correction,flist_img_to_filter, \
                                illumination_function=Illumination_norm,\
                                filtered_png_img_gene_dirs=filtered_png_img_gene_dirs,\
                                filtered_img_gene_dirs =filtered_img_gene_dirs,\
                                counting_gene_dirs=counting_gene_dirs,plane_keep=flt_rawcnt_config['plane_keep'], \
                                min_distance=flt_rawcnt_config['min_distance'], stringency=flt_rawcnt_config['stringency'],\
                                skip_genes_counting=flt_rawcnt_config['skip_genes_counting'],skip_tags_counting=flt_rawcnt_config['skip_tags_counting'])
                client.gather(futures_processes)
               

        else:
            for gene in hybridizations_infos[hybridization].keys():
                flist_img_to_filter=glob.glob(hyb_dir+processing_hyb+'_tmp/'+processing_hyb+'_'+gene+'_tmp/*.npy')
                # filtering
                logger.debug('Filtering without illumination correction %s',gene)

                futures_processes=client.map(counting.filtering_and_counting,flist_img_to_filter, \
                                        filtered_png_img_gene_dirs=filtered_png_img_gene_dirs, \
                                        filtered_img_gene_dirs=filtered_img_gene_dirs, \
                                        counting_gene_dirs=counting_gene_dirs, \
                                        plane_keep=flt_rawcnt_config['plane_keep'], min_distance=flt_rawcnt_config['min_distance'],\
                                        stringency=flt_rawcnt_config['stringency'],\
                                        skip_genes_counting=flt_rawcnt_config['skip_genes_counting'],skip_tags_counting=flt_rawcnt_config['skip_tags_counting'])

                client.gather(futures_processes)
                
        # ---------------------------------------------------------------------
        
        # # ----------------- COMBINE THE FILTERED DATA IN .ppf.hdf5 ------------------------
        # # Combine the filter data in one single .ppf for each hybridization
        # # This step will run in serial mode and will not need to shuffle data
        # #  between cores because everything is on the common file system

        # logger.debug('Create .ppf.hdf5 file')

        # # Create the ppf.hdf5 file that contains the filtered data in uint16
        # preprocessing_file_path = hdf5_utils.hdf5_create_preprocessing_file(hybridizations_infos,processing_hyb,
        #                                 hybridization,flt_rawcnt_config['analysis_name'], hyb_dir,converted_positions,image_properties)

        # logger.debug('Write the .npy filtered files into the .ppf file')
        # # Load and write the .npy tmp images into the hdf5 file

        # # open the hdf5 file
        # with h5py.File(preprocessing_file_path) as f_hdl:
        #     # Loop through each gene
        #     for gene in hybridizations_infos[hybridization].keys():

        #         logger.debug('Writing %s images in .ppf.hdf5',gene)
        #         # list of the files to transfer
        #         filtered_gene_dir = [fdir for fdir in filtered_img_gene_dirs if gene in fdir][0]
        #         filtered_files_list = glob.glob(filtered_gene_dir+'*.npy')

        #         # loop through the list of file
        #         for f_file in filtered_files_list:
        #             pos = f_file.split('/')[-1].split('_')[-1].split('.')[0]
        #             f_hdl[gene]['FilteredData'][pos][:] =np.load(f_file)
        #             f_hdl.flush()
        
        # # ---------------------------------------------------------------------
        
        # ----------------- STITCHING ------------------------
        # Load the stitching parameters from the .yaml file

        # Stitch the image in 2D or 3D (3D need more work/testing)
        nr_dim = flt_rawcnt_config['nr_dim']

        # Estimated overlapping between images according to the Nikon software
        est_overlap = image_properties['Overlapping_percentage']

        # Number of peaks to use for the alignment
        nr_peaks = flt_rawcnt_config['nr_peaks']

        # Determine if the coords need to be flipped

        y_flip = flt_rawcnt_config['y_flip']

        # Method to use for blending
        # can be 'linear' or 'non linear'
        # The methods that performs the best is the 'non linear'

        blend = flt_rawcnt_config['blend']

        # Reference gene for stitching
        reference_gene = flt_rawcnt_config['reference_gene']

        pixel_size = image_properties['PixelSize']

        # Get the list of the filtered files of the reference gene
        filtered_gene_dir = [gene_dir for gene_dir in filtered_img_gene_dirs if reference_gene in gene_dir][0]
        filtered_files_list = glob.glob(filtered_gene_dir+'*.npy')

        # Create pointer of the hdf5 file that will store the stitched reference image
        # for the current hybridization
        # Writing
        tile_file_base_name = flt_rawcnt_config['analysis_name']+'_'+ processing_hyb
        data_name   = (tile_file_base_name
                        + '_' + reference_gene
                        + '_stitching_data')

        stitching_file_name = tile_file_base_name + '.sf.hdf5'
        stitching_file= h5py.File(hyb_dir+stitching_file_name,'w',libver='latest')  # replace with 'a' as soon as you fix the error


        # Determine the tiles organization
        tiles, contig_tuples, nr_pixels, z_count, micData = stitching.get_pairwise_input_npy(image_properties,converted_positions, hybridization,
                                est_overlap = est_overlap, y_flip = False, nr_dim = 2)



        # Align the tiles 
        futures_processes=client.map(pairwisesingle.align_single_pair_npy,contig_tuples,
                                    filtered_files_list=filtered_files_list,micData=micData, 
                                nr_peaks=nr_peaks)

        # Gather the futures
        data = client.gather(futures_processes)


        # In this case the order of the returned contingency tuples is with
        # the order of the input contig_tuples

        # P_all = [el for data_single in data for el in data_single[0]]
        P_all =[data_single[0] for data_single in data ]
        P_all = np.array(P_all)
        P_all = P_all.flat[:]
        covs_all = [data_single[1] for data_single in data]
        alignment = {'P': P_all,
                    'covs': covs_all}


        # Calculates a shift in global coordinates for each tile (global
        # alignment) and then applies these shifts to the  corner coordinates
        # of each tile and returns and saves these shifted corner coordinates.
        joining = stitching.get_place_tile_input(hyb_dir, tiles, contig_tuples,
                                                    micData, nr_pixels, z_count,
                                                    alignment, data_name,
                                                    nr_dim=nr_dim)

        # Create the hdf5 file structure
        stitched_group, linear_blending, blend =  hdf5preparation.create_structures_hdf5_stitched_ref_gene_file_npy(stitching_file, joining, nr_pixels,
                                        reference_gene, blend = 'non linear')

        # Fill the hdf5 containing the stitched image with empty data and
        # create the blending mask
        stitched_group['final_image'][:]= np.zeros(joining['final_image_shape'],dtype=np.float64)
        if blend is not None:
            # make mask
            stitched_group['blending_mask'][:] = np.zeros(joining['final_image_shape'][-2:],dtype=np.float64)
            tilejoining.make_mask(joining, nr_pixels, stitched_group['blending_mask'])

            
        # Create the subdirectory used to save the blended tiles
        suffix = 'blended_tiles'
        blended_tiles_directory = utils.create_single_directory(hyb_dir,reference_gene, hybridization,processing_hyb,suffix,add_slash,
                                        analysis_name=flt_rawcnt_config['analysis_name'])

        # Get the directory with the filtered npy images of the reference_gene to use for stitching
        stitching_files_dir = [npy_dir for npy_dir in filtered_img_gene_dirs if reference_gene in npy_dir][0]


        # Create the tmp directory where to save the masks
        suffix = 'masks'
        masked_tiles_directory = utils.create_single_directory(hyb_dir,reference_gene, hybridization,processing_hyb,suffix,add_slash,
                                        analysis_name=flt_rawcnt_config['analysis_name'])

        # Create and save the mask files
        for corn_value,corner_coords in joining['corner_list']:
            if not(np.isnan(corner_coords[0])):
                cur_mask = stitched_group['blending_mask'][int(corner_coords[0]):int(corner_coords[0]) + int(nr_pixels),
                                    int(corner_coords[1]):int(corner_coords[1]) + int(nr_pixels)]

                fname = masked_tiles_directory + flt_rawcnt_config['analysis_name'] +'_'+processing_hyb+'_'+reference_gene+'_masks_joining_pos_'+str(corn_value)
                np.save(fname,cur_mask)


        # Blend all the tiles and save them in a directory
        futures_processes = client.map(tilejoining.generate_blended_tile_npy,joining['corner_list'],
                                    stitching_files_dir = stitching_files_dir,
                                    blended_tiles_directory = blended_tiles_directory,
                                    masked_tiles_directory = masked_tiles_directory,
                                    analysis_name = flt_rawcnt_config['analysis_name'],
                                    processing_hyb = processing_hyb,reference_gene = reference_gene,
                                    micData = micData,tiles = tiles,nr_pixels=nr_pixels,
                                    linear_blending=linear_blending)



        _ = client.gather(futures_processes)


        # Write the stitched image
        tilejoining.make_final_image_npy(joining, stitching_file, blended_tiles_directory, tiles,reference_gene, nr_pixels)

        # close the hdf5 file
        stitching_file.close()


        # Delete the directories with blended tiles and masks
        shutil.rmtree(blended_tiles_directory)
        shutil.rmtree(masked_tiles_directory)

        # ----------------- DELETE FILES ------------------------
        # Don't delete the *.npy files here because can be used to 
        # create the final images using the apply stitching related function    









    client.close()
    

if __name__ == '__main__':
    preprocessing_script()


import numpy as np
from skimage import filters,img_as_float
from scipy import ndimage as nd

def nuclei_filtering(img_stack):
    """
    This function remove the background from the nuclei 
    For the sigma I seleced a value quite bigger than
    the nuclei size in order to remove them from the 
    image. I used what showed on the gaussian filter code page and on this
    link on stackoverflow: 
    http://stackoverflow.com/questions/25216382/gaussian-filter-in-scipy

    Arguments
    -----------

    img_stack: np.array float64
        3D numpy array with the image
    
    Returns
    -----------

    filtered_image: np.array float64 
        2D flattened image 

    """

    # Clean the image from the background
    img_stack = img_stack-filters.gaussian(img_stack,sigma=(2,100,100))
    
    # Remove the negative values        
    img_stack[img_stack<0] = 0

    # Flatten the image
    filtered_image = np.amax(img_stack,axis=0)

    return filtered_image


def smFISH_filtering(img_stack):
    """
    This function remove the background from the smFISH and enhance the dots.
    
    Arguments
    -----------

    img_stack: np.array float64
        3D numpy array with the image
    
    Returns
    -----------

    filtered_image: np.array float64 
        2D flattened image 
    """

    # Use a gaussian with kernel bigger than the dots to estimate the background
    # and remove it from the image
    img_stack = img_stack-filters.gaussian(img_stack,sigma=(1,8,8))
    img_stack[img_stack<0] = 0

    # Enhance the dots
    img_stack = nd.gaussian_laplace(img_stack,sigma=(0.2,0.5,0.5))

    img_stack = -img_stack # the peaks are negative so invert the signal
    img_stack[img_stack<0] = 0 # All negative values set to zero 

    # Flatten the image
    filtered_image = np.amax(img_stack,axis=0)

    return filtered_image



import numpy as np
from sympy import Point, Line, Segment
from skimage import filters,io,img_as_float,exposure,morphology,segmentation,measure,feature
from scipy import ndimage as nd
import itertools


def image_chunks_calculator(dimension,chunk_size):
    """
    Helper function to calculate the size of the chunks created according
    the length of the vector and the chunk size.

    Parameters:
    -----------

    dimension: int
        Length of the vector to Chunk
    chunkSize: int 
        Dimension of the Chunks

    Returns:
    -----------

    chunks_sizes: np.array 
        Array of the sizes of the created chunks. It deals with conditions 
        when the expected chunks size do not fit an even number of times in the 
        dimension
    """
    number_even_chunks=int(dimension//chunk_size)
    total_size_even_chunks=number_even_chunks*chunk_size
    odd_tile_size=dimension-total_size_even_chunks
    chunk_sizes=[]
    chunks_sizes=list(np.repeat(chunk_size,number_even_chunks-1))
    if odd_tile_size < chunk_size:
        chunks_sizes.append(chunk_size+odd_tile_size)
    else:
        chunks_sizes.append(odd_tile_size)
    return tuple(chunks_sizes)



def image_chunking(img,chunk_size,percent_padding):
    """
    Function used to generate the coords of the images according to the
    chunking 

    Parameters:
    -----------

    stitched_file_hdl: pointer 
        Handle of the hdf5 file that contains the stitched image.
    gene: str 
        Processed gene ('Nuclei').
    PercentPadding: float 
        Percent of overlapping between the different images (Ex. 0.2).
    ChunkSize: int 
        Dimension of the Chunks.

    Returns:
    -----------

    Coords_Chunks_list: list 
        List of np.array with the coords of the images without padding
    Coords_Padded_Chunks_list: list 
        List of np.array with the coords of the images with padding
    
    Notes:
    ------

    For both lists each np.array contains the coords in the following order:
    [row_tl,row_br,col_tl,col_br]

    """
#     img_r,img_c=stitched_file_hdl[gene][FolderLevel]['final_image'].shape
    
    img_r,img_c=img.shape
    pixel_padding = int(chunk_size*percent_padding)


    # Calculate the size of the chunks
    r_chunks_size = image_chunks_calculator(img_r,chunk_size)
    c_chunks_size = image_chunks_calculator(img_c,chunk_size)
    # Calculate the total numbers of chunks
    nr_chunks = len(r_chunks_size)
    nc_chunks = len(c_chunks_size)


    # Coords top left corner (tl)
    r_coords_tl = np.arange(0,chunk_size*(nr_chunks),chunk_size)
    c_coords_tl = np.arange(0,chunk_size*(nc_chunks),chunk_size)
    # Coords of all the tl in the image
    r_coords_tl_all,c_coords_tl_all = np.meshgrid(r_coords_tl,c_coords_tl,indexing='ij')

    # Calculate all the br coords
    r_coords_br_all = r_coords_tl_all.copy()
    c_coords_br_all = c_coords_tl_all.copy()

    for c in np.arange(0,r_coords_tl_all.shape[1]):
        r_coords_br_all[:,c] = r_coords_br_all[:,c]+r_chunks_size

    for r in np.arange(0,r_coords_tl_all.shape[0]):
         c_coords_br_all[r,:] = c_coords_br_all[r,:]+c_chunks_size

    # Calculate the padded coords
    r_coords_tl_all_padded = r_coords_tl_all-pixel_padding
    c_coords_tl_all_padded = c_coords_tl_all-pixel_padding
    r_coords_br_all_padded = r_coords_br_all+pixel_padding
    c_coords_br_all_padded = c_coords_br_all+pixel_padding

    # Correct for coords out of the image (where tl<0,br>Img.shape)
    r_coords_tl_all_padded[r_coords_tl_all_padded<0] = r_coords_tl_all[r_coords_tl_all_padded<0]
    c_coords_tl_all_padded[c_coords_tl_all_padded<0] = c_coords_tl_all[c_coords_tl_all_padded<0]
    r_coords_br_all_padded[r_coords_br_all_padded>img_r] = r_coords_br_all[r_coords_br_all_padded>img_r]
    c_coords_br_all_padded[c_coords_br_all_padded>img_c] = c_coords_br_all[c_coords_br_all_padded>img_c]

    # The coords list are generated as:
    # row_tl,row_br,col_tl,col_br

    # Create a list for the non padded coords
    Coords_Chunks_list = list()
    for r in np.arange(0,r_coords_tl_all.shape[0]):
        for c in np.arange(0,r_coords_tl_all.shape[1]):
            Coords_Chunks_list.append((np.array([r_coords_tl_all[r][c],\
                                                 r_coords_br_all[r][c],\
                                                 c_coords_tl_all[r][c],\
                                                 c_coords_br_all[r][c]])))

    # Create a list for the padded coords
    Coords_Padded_Chunks_list = list()
    for r in np.arange(0,r_coords_tl_all_padded.shape[0]):
        for c in np.arange(0,r_coords_tl_all_padded.shape[1]):
            Coords_Padded_Chunks_list.append(np.array([r_coords_tl_all_padded[r][c],\
                                                       r_coords_br_all_padded[r][c],\
                                                       c_coords_tl_all_padded[r][c],\
                                                       c_coords_br_all_padded[r][c]]))

    return nr_chunks,nc_chunks,Coords_Chunks_list, Coords_Padded_Chunks_list,r_coords_tl_all_padded,\
            c_coords_tl_all_padded,r_coords_br_all_padded,c_coords_br_all_padded

def polyT_segmentation(Img, min_object_size, min_distance,disk_radium_rank_filer,threshold_rel,trimming):
    
    """
    Function that runs the polyT segmentation on an image

    Parameters:
    -----------

    Img: np.array
        2D image to be segmented
    min_object_size: int
        Minimum size of a connected component in order to be considered a cell.
    min_distance: int
        Minimum distance between peaks for the creation of the markers for the
        first round of watershed.
    disk_radium_rank_filer: int
        Size of the filter for preprocessing the image in order to calculate the
        Thr.
    threshold_rel: float
        relative threshold between the markers for the first round of watershed.
    trimming: int 
        Define the trimming point for the selection of the Thr for image 
        processing.


    Returns:
    --------

    Objects_dict: dict
        Dictionary with the labels and their coords.

    """


    Img = img_as_float(Img)
    Hist=exposure.histogram(Img)
    ThrArray=Hist[1]
    SelectedThr = ThrArray[trimming]
    
    # Consider the case with a black image
    Objects_dict=dict()
    if Img.max()==0:
        Objects_dict={}
    else:

    #     SelectedThr=rescaling_calculation(Img,relax)
        Img=exposure.rescale_intensity(Img,in_range=(0,SelectedThr))
        Img_Thr=filters.rank.maximum(Img,morphology.disk(disk_radium_rank_filer))
        # If ranking filter remove all the objects the image is made of all 0
        # and otsu won't work

        if Img_Thr.max()==0:
            Objects_dict={}
        else:

            Threshold=filters.threshold_otsu(Img_Thr)
            Img_Thr=Img_Thr>Threshold

            # Fill holes in the binarized objects
            ImgFill=nd.morphology.binary_fill_holes(Img_Thr)
            # Binary opening
            Img_Thr_op=nd.binary_opening(ImgFill,structure=morphology.disk(2))
            # Remove small object
            ImgRemSmall=morphology.remove_small_objects(Img_Thr_op,min_size=min_object_size)

            # Determine if by removing the small objects you fully cleared the image
            Check = np.asarray(ImgRemSmall, dtype=int)
            Check = Check.sum()
            if Check>0:
                # Identify the objects 
                labels_peak=measure.label(ImgRemSmall,connectivity=2,background=0)

                # Create markers for the first round of watersher
                distance=nd.distance_transform_edt(Img*ImgRemSmall)

                local_maxi = feature.peak_local_max(distance,min_distance=min_distance,threshold_rel=threshold_rel,\
                                                    indices=False,labels=labels_peak,exclude_border = False)

                # Take care of the case where there is an object but no local max and
                # cannot determine the number of values
                if local_maxi.max()>0:
                    marker=markers = nd.label(local_maxi)[0]
                    # Remember that background is 0
                    Objects_wt_one = morphology.watershed(-distance, marker, mask=ImgRemSmall)
                    # no need for intensity...need to be recalculate lated
                    #Objects_wt_one_prop=measure.regionprops(Objects_wt_one,intensity_image=Img)
                    Objects_wt_one_prop=measure.regionprops(Objects_wt_one)

                    # Select only newly created objects above minsize 
                    for obj in Objects_wt_one_prop:
                        if obj.area> min_object_size:
                            # Return only the labels and the corresponding coords
                            Objects_dict[obj.label]=obj.coords
                else:
                    Objects_dict={}
            else:
                Objects_dict={}
    return Objects_dict

def OverlappingCouples(chunk_coords,TotalDataDict):
    """
    Calculate all the intersecting objects
    
    Parameters:
    -----------

    chunk_coords: list 
        List of np.array with the coords of the images with padding
    TotalDataDict: dict
        Dictionary with the labels and the coords of the identified objects
        after merging the result of the segmentation on the different image
        chunks.

    Returns:
    --------

    Intersecting_cpl: list
        List of tuples with all the overlapping labels
        [(Obj1,Obj12),(Obj2, Obj3)].

    """
    # Filter out the objects not in the overlapping region
    ObjectsOverlap_list=list()
    Intersecting_cpl=list()
    Intersecting_cpl_idx=list()
    Intersection=list()
    for lab in TotalDataDict.keys():
        if ((TotalDataDict[lab][:,0] >= chunk_coords[0]).any() and
            (TotalDataDict[lab][:,0] <= chunk_coords[1]).any() and
            (TotalDataDict[lab][:,1] >= chunk_coords[2]).any() and
            (TotalDataDict[lab][:,1] <= chunk_coords[3]).any()):
                ObjectsOverlap_list.append(lab)
    
    # Remove possible duplicates
    ObjectsOverlap_list=list(set(ObjectsOverlap_list))
    
    # Create all possible combinations of overlapping objects
    combinations=list(itertools.combinations(ObjectsOverlap_list,2))
    
    # For all the couples of objects determine which one are overlapping 
    for couple in combinations:
        Result=not set(map(tuple, TotalDataDict[couple[0]])).isdisjoint(map(tuple, TotalDataDict[couple[1]]))
        Intersection.append(Result)
    if (np.array(Intersection)).any():
        [Intersecting_cpl_idx.append(i) for i,x in enumerate(Intersection) if x]
        [Intersecting_cpl.append(combinations[cp]) for cp in Intersecting_cpl_idx]
    
    return Intersecting_cpl

# Function for calculating the intersection. Recalculate the coords
# Output the coords and the labels to keep and to get rid off
# in this case I merge the coords but in the future I will merge then recalculate
# the segmentation
def IntersectionCouples(couple,TotalDataDict):

    """
    Function to consolidate the overlapping couples. Identify if there are
    multiple overlapping, merge the coords together and keep only one label
    
    Parameters:
    -----------
    couple: list
        List of tuples with the labels of the overlapping objects. If contains
        more than one repeat of the same object, the repeats are removed.
    TotalDataDict: dict
        Dictionary with the labels and the coords of the identified objects
        after merging the result of the segmentation on the different image
        chunks.   

    Returns:
    --------
    SaveLab: list
        List of saved labels.
    RemoveLabs: list
        List of the remove labels.
    NewCoords: list
        Merged obj coords.

    """
    # If there are more then one combination in the couple
    if len(couple)>1:
        # Remove the duplicated
        labs=list()
        [labs.append(element) for cpl in couple for element in cpl]
        labs=list(set(labs))
    else:
        labs=list(couple[0])
    # Merge the coords
    NewCoords=TotalDataDict[labs[0]]
    for lab in labs[1:]:
        NewCoords=np.vstack((NewCoords,TotalDataDict[lab]))

    # Remove duplicates
    Tp_NewCoord=[tuple(x) for x in NewCoords]
    NewCoords=np.array(list(set(Tp_NewCoord)))

    TotalDataDict[labs[0]]=NewCoords
    
    SaveLab=labs[0]
    RemoveLabs=labs.copy()
    RemoveLabs.remove(SaveLab)
    
    return SaveLab, RemoveLabs,NewCoords 

def obj_properties_calculator(obj):

    """
    Calculate the morphological properties of the object (connected component).
    
    Parameters:
    -----------

    obj: dict
        Dictionary with the object label and the coords.

    Returns:
    --------
    
    obj_prop_dict: dict
        Dictionary with the calculated properties of the object


    """
    
    # Create the obj properties dictionary
    obj_prop_dict = dict()
    
    # Define ID and coords of the obj
    obj_id = obj[0] 
    obj_coords = obj[1]
    
    row_Coords_obj=obj_coords[:,0]
    col_Coords_obj=obj_coords[:,1]
    
    # Calculate the coords of the bounding box to use to crop the ROI
    # Added +1 because in the cropping python will not consider the end of the interval
    bb_row_max=row_Coords_obj.max()+1
    bb_row_min=row_Coords_obj.min()
    bb_col_max=col_Coords_obj.max()+1
    bb_col_min=col_Coords_obj.min()
    
    
    # Normalize the coords of the obj
    row_Coords_obj_norm=row_Coords_obj-bb_row_min
    col_Coords_obj_norm=col_Coords_obj-bb_col_min
    
    
    # Bounding box
    bb_coords = (bb_row_min,bb_row_max,bb_col_min,bb_col_max)
    
    # obj_coords recalculated
    obj_coords = (row_Coords_obj_norm,col_Coords_obj_norm)
    
    
    # Create object image for properties
    Img=np.zeros([bb_coords[1]-bb_coords[0],bb_coords[3]-bb_coords[2]],dtype=np.uint8)
    Img[obj_coords[0],obj_coords[1]]=1

    # calculate the object properties
    obj_prop=measure.regionprops(Img)

    # Collect only the needed properties
    obj_area=obj_prop[0].area
    obj_centroid=obj_prop[0].centroid
    obj_convex_area=obj_prop[0].convex_area
    obj_filled_area=obj_prop[0].filled_area
    obj_major_axis_length=obj_prop[0].major_axis_length
    obj_minor_axis_length=obj_prop[0].minor_axis_length
    obj_perimeter=obj_prop[0].perimeter

    # Adjust coords of the centroid
    obj_centroid=[obj_centroid[0]+bb_coords[0],obj_centroid[1]+bb_coords[2]]
  
    obj_prop_dict[obj_id]={'obj_area':obj_area,'obj_centroid':obj_centroid,'obj_convex_area':obj_convex_area,\
                        'obj_filled_area':obj_filled_area,'obj_major_axis_length':obj_major_axis_length,\
                        'obj_minor_axis_length':obj_minor_axis_length,'obj_perimeter':obj_perimeter,
                        'obj_coords':obj[1]}
    
    return obj_prop_dict



import numpy as np
import numpy.ma as ma
import logging

#Own imports
from . import inout

class MicroscopeData:
    """Retrieves and stores the tile set and corner coordinates of each
    tile from raw coordinate input.

    Parameters:
    -----------

    y_flip: bool
        It is assumed that the microscope
        coordinates have their origin in right top of the image.
        During the stitching we use the left, top as the origin.
        Therefore the x-coordinates will be inverted by the
        normalization. But not the y-coords.
        The y_flip variable is designed for the cases where the
        microscope sequence is inverted in the y-direction. When
        set to True the y-coordinates will also be inverted
        before determining the tile set.
    x_coords: np.array
        Array of x coordinates, will be loaded
        as raw microscope coordinates in um by init() and
        converted to pixels and normalized to start at zero by
        normalize_coords().
    y_coords: np.array
        Array of y coordinates, will be loaded
        as raw microscope coordinates in um by init() and
        converted to pixels and normalized to start at zero by
        normalize_coords().
    z_coords: np.array
        Array of z coordinates, will be loaded
        as raw microscope coordinates in um by init() and
        converted to pixels and normalized to start at zero by
        normalize_coords().
    tile_nr: np.array
        The numbers of the tiles as found in coord_data.
    tile_set: np.array
        Array with the same shape as the
        tile set, values indicate the index at which the
        corresponing tile is found in tiles and tile_nr.
    running_av: float
        The estimated overlap between two
        neighbouring tiles in pixels, adapted for the tiles we
        already know the placement and overlap of.
    logger: logger
        logger instance
    """

    def __init__(self, coord_data, y_flip, nr_dim):
        """ Create looger, read-in the coordinates and tile numbers,
        create empty tile set.

        The tile numbers refer to the number in the file name of the
        image, they identify which image file belongs to which tile.

        Parameters:
        -----------

        coord_data: dict
            Dictionary with for each tile a list with
            x, y and z coordinate, where the key is the tile
            number.
            utils.experimental_metadata_parser() outputs such a
            dictionary when reading a file with microscope
            coordinates.
        y_flip: bool
            Designed for the cases where the
            microscope sequence is inverted in the
            y-direction. When set to True the y-coordinates will
            be inverted before determining the tile set. The
            x-direction is assumed to be always inverted.
        nr_dim: int
            Valid values: 2 and 3. The number of
            dimension the loaded tiles and stitched image will have.
        """
        #Create logger
        self.logger = logging.getLogger(__name__)

        #Set flag for y flip:
        self.y_flip = y_flip

        #Extract coordinates
        self.x_coords   = []
        self.y_coords   = []
        self.z_coords   = []
        self.tile_nr    = []
        # Do not substract 1 anymore to compensate for difference in
        # filenames and image numbers in coordinates file
        self.logger.debug('Reading coord_data:')
        for key, row in coord_data.items():
            self.logger.debug('key: {} row: {}'.format(key, row))
            self.x_coords.append(row[0])
            self.y_coords.append(row[1])
            self.z_coords.append(row[2])
            self.tile_nr.append(int(key))
        #Make sure everything is np arrays
        self.x_coords   = np.array(self.x_coords)
        self.y_coords   = np.array(self.y_coords)
        self.z_coords   = np.array(self.z_coords)
        self.tile_nr    = np.array(self.tile_nr)
        self.logger.debug(("Parsed microscope data:\n tile_nr: {} \n "
                          + "x: {} \n y: {}")
                          .format(self.tile_nr, self.x_coords,
                                  self.y_coords))
        #self.logger.debug("x, y coord and tile nr: {} {} {}"
        #                        .format(self.x_coords, self.y_coords, self.tile_nr))
        #Init tile set
        self.tile_set = []


    def __getstate__(self):
        """Causes objects to be pickled without the logger attribute"""
        class_dict = dict(self.__dict__)
        del class_dict['logger']
        return class_dict

    def __setstate__(self, state):
        """Restores the logger atribute when unpickling the object"""
        # Restore instance attributes (i.e., filename and lineno).
        self.__dict__.update(state)
        # Recreate the logger file, that has not been pickled
        self.logger = logging.getLogger(__name__)

    def normalize_coords(self, pixel_size):
        """Normalize the coordinates and put the origin in upper left
        corner.
        Takes list with absolute index, x, y values from microscope
        and normalizes, converts and inverts them to get the coordinates
        in pixels with the origin in left upper corner.

        NB:In the new microscope setup the origin is in the bottom right 
        corner.


        Parameters:
        -----------

        pixel_size: float
            The size of one pixel in the microscope
            image, used to convert the microscope coordinates to pixels.
        """
        self.logger.info("Normalizing coordinates")
        # Convert to pixels
        self.x_coords /= pixel_size
        self.y_coords /= pixel_size
        # Commented out because in the new system the origin is in the bottom right
        # Flip the coordinate system to get tiles in right order (0,0 is left top)

        # OLD SYSTEM

        # self.x_coords *= -1
        # if self.y_flip:
        #     self.y_coords *= -1
        # # Normalization to get postive coordinates
        # x_min = np.amin(self.x_coords)
        # self.x_coords -= x_min
        # y_min = np.amin(self.y_coords)
        # self.y_coords -= y_min

        # self.x_coords *=1
        # self.y_coords *=1
        # if self.y_flip:
        #     self.y_coords *= -1
        
        # # NEW SYSTEM

        # Normalization to get postive coordinates
        x_min = np.amin(self.x_coords)
        self.x_coords -= x_min
        y_min = np.amin(self.y_coords)
        self.y_coords -= y_min

        x_max=np.amax(self.x_coords)
        self.x_coords -= x_max
        self.x_coords=np.abs(self.x_coords)

        y_max=np.amax(self.y_coords)
        self.y_coords -= y_max
        self.y_coords=np.abs(self.y_coords)


        self.logger.debug(('Normalized microscope data:\n tile_nr: {} '
                          + '\n x: {} \n y: {}')
                          .format(self.tile_nr, self.x_coords,
                                  self.y_coords))


    def make_tile_set(self, est_x_tol, nr_pixels, row_tol = None):
        """Based on the coordinates find a tile set.

        Use the coordinates to produce a tile set that has the shape
        of the grid the tiles should be placed on and for each tile in
        this grid gives the index for the tile_nr, x_coords and
        y_coords.
        Plots the coordinates when plot_avaible == True in inout.

        Parameters:
        -----------

        est_x_tol: int
            The estimated difference in the x
            direction between the corners of two
            neighbouring tiles in pixels
        nr_pixels: int
            Needed to estimate the distance between two separate tiles
        row_tol: int
            Row tolerance in pixels: The distance
            between y coordinates above which a tile is
            considered to belong to a new row. Default is None,
            which leads to 0.5 * nr_pixels.
        """
        self.logger.info("Making tile set")
        # If row_tolerance in not passed; set the row tolerance
        # according to number of pixels:
        if row_tol is None:
            row_tol = 0.5 * nr_pixels
        self.logger.debug("Row tolereance: {}".format(row_tol))
        # Pre-sort y-coords
        coord_inds          = np.arange(0,len(self.tile_nr))
        sorting_inds        = np.argsort(self.y_coords)
        sorted_coord_inds   = coord_inds[sorting_inds]
        sorted_y            = self.y_coords[sorting_inds]

        cur_row = [sorted_coord_inds[0]]
        # Sort into rows, according to y-coordinates:
        self.logger.debug("Finding rows...")
        for i in range(1,len(sorted_y)):
            # Check the difference with the neighbour
            if abs(sorted_y[i - 1] - sorted_y[i]) > row_tol:
                self.tile_set.append(list(cur_row))
                self.logger.debug("Added row: {} ".format(cur_row))
                cur_row = [sorted_coord_inds[i]]
            else:
                cur_row.append(sorted_coord_inds[i])
        # Dump leftover in last row of tile_set
        self.tile_set.append(list(cur_row))
        self.logger.debug('Added last row: {}'.format(cur_row))

        # Sort each row according to x-coordinates
        # Initialize
        self.logger.debug("Sorting within rows...")
        self.running_av  = est_x_tol
        x_max       = max(self.x_coords) + nr_pixels
        nr_col      = int(np.around(x_max / est_x_tol))
        self.logger.debug(("Estimated x-size of final picture: {}, "
                          + "Number of expected collumns in tile "
                          + "set: {}")
                          .format(x_max, nr_col))
        av_counter  = 1
        # Sort each row in tile_set
        for i in range(len(self.tile_set)):
            # Convert to np-array
            self.tile_set[i] = np.array(self.tile_set[i])
            # Sort row
            sorting_inds = np.argsort(self.x_coords[self.tile_set[i]])
            self.logger.debug(("Row {}. "
                              + "Sorted x-coordinates:\n {}")
                              .format(i, self.x_coords[self.tile_set[
                                i]][sorting_inds]))

            # Local variable for easy adjustmet of sorted tiles
            sorted_tiles = self.tile_set[i][sorting_inds]
            #self.logger.debug("Current sorted tiles {}".format(
            # sorted_tiles))

            # Make local copy with 0 appended, for easy adjustment of sorted x coordinates
            sorted_x = np.concatenate(([0.0], self.x_coords[self.tile_set[i]][sorting_inds]))

            # Check for missing tiles:
            # Get distance between the tiles
            diff_x = abs(sorted_x[1:] - sorted_x[:-1])
            self.logger.debug("Row {}. Distance between tiles:\n {}"
                              .format(i, diff_x))
            # Update the average distance between directly adjacent tiles
            neighbouring_tiles = np.argwhere(diff_x[1:] < nr_pixels)
            if neighbouring_tiles.any():
                #self.logger.debug("running av calculated with {}".format(diff_x[1:][neighbouring_tiles]))
                #self.logger.debug("running av calculated with {}".format(np.mean(diff_x[1:][neighbouring_tiles])))
                av_counter = 2
                self.running_av = (self.running_av
                                    + np.mean(diff_x[1:][neighbouring_tiles])) \
                                    / av_counter
            self.logger.debug(("Row {}. Running average of estimated "
                              + "overlap: {}")
                              .format(i, self.running_av))

            # Find missing tiles (ignore the first tile, this one is checked later)
            missing_tiles = np.argwhere(diff_x[1:] > nr_pixels)
            missing_tiles = [tile[0] + 1 for tile in missing_tiles]
            self.logger.debug("Row {}. Missing tiles: {}".format(i,
                missing_tiles))

            # Check how many tiles are missing and insert substituting value: -1
            for diff_ind in missing_tiles:
                #self.logger.debug("Difference {}".format(diff_x[diff_ind]))
                nr_missing_tiles = int(np.around(diff_x[diff_ind] / self.running_av) - 1)
                ins_value = np.full(nr_missing_tiles, -1, dtype = int)
                #self.logger.debug("inserting missing tile at index {}".format(diff_ind))
                sorted_tiles = np.insert(sorted_tiles, diff_ind, ins_value)
                #self.logger.debug("inserting missing tile, tiles: {}".format(sorted_tiles))

            # Check for missing tiles at the start insert substituting value: -1
            if diff_x[0] > row_tol:
                nr_missing_tiles = int(np.around(diff_x[0] / self.running_av))
                ins_value = np.full(nr_missing_tiles, -1, dtype = int)
                sorted_tiles = np.insert(sorted_tiles, 0, ins_value)

            # Check if this row is as long as the rest, if it not, assume
            # that tiles are missing at the end
            # and insert substituting value: -1
            if len(sorted_tiles) < nr_col:
                nr_missing_tiles = nr_col - len(sorted_tiles)
                ins_value = np.full(nr_missing_tiles, -1, dtype = int)
                sorted_tiles = np.append(sorted_tiles, ins_value)

            # Mask the substituting value
            sorted_tiles = ma.masked_equal(sorted_tiles, -1)
            self.logger.debug("Row {}. Current sorted tiles {}".format(
                i, sorted_tiles))
            # Add to tile set:
            self.tile_set[i]  = sorted_tiles


        # Mask the tile set
        self.tile_set = ma.array(self.tile_set)
        #self.tile_set = np.array(self.tile_set)

        # Logging and plot to check
        self.logger.info("Tile set:\n {}".format(self.tile_set))
        self.logger.info("Tile set shape: {}".format(
            self.tile_set.shape))
        self.logger.info("Tile numbers:\n {}".format( self.tile_nr))

        inout.plot_coordinates(self, invert_yaxis=False)


    def check_tile_set(self, est_x_tol, nr_pixels, row_tol = None):
        """Check if the estimated overlap between tiles is close enough
        to the running average.

        Parameters:
        -----------

        est_x_tol: int
            The distance between two tiles in the
            x-direction as estimated before reading all coordinates.
        nr_pixels: int
            The the width of the tile in pixels.
        row_tol: int
            The distance between y coordinates
            above which a tile is considered to belong to a
            new row, this value is passed directly to
            make_tile_set. (default = None)
        """
        x_max       = max(self.x_coords)
        nr_col      = int(np.around(x_max / est_x_tol))
        if ((abs(est_x_tol - self.running_av) > 1000)
            or not(self.tile_set.shape[1] == nr_col)):
                self.make_tile_set(self.running_av, nr_pixels,
                                    row_tol = row_tol)



# Module containing functions written for the in and output of tile microcope images
from skimage import io

# import matplotlib.pyplot as plt
# import visvis as vv
# plt_available = True
plt_available = False

import glob as glob
import pickle as pickle
import numpy as np
import logging
import csv
import ast
import re

def extract_hdf5_refs(hdf5_group, image_inds):
    """Make a list of the dataset is hdf5_group.

    When image_inds is not empty this function returns only the
    data sets corresponding to the numbers in image_inds.

    Parameters:
    -----------

    hdf5_group: pointer
        HDF5 group reference. Reference to the group
        from which we should extract the dataset names.
        The data sets are expected to be numbered.
    image_inds: iterable
        Iterable containing ints. Should contain the
        names of the datasets that will be selected. This
        list could also contain strings, as long as it's
        contents reflect a subset of dataset names.
    
    Returns:
    --------

    tile_refs: list
        List of strings. The references to the data sets.
    """
    # Get logger
    logger = logging.getLogger(__name__)
    # Get the full human readable reference to hdf5_group.
    group_root_name = hdf5_group.name + '/'
    if image_inds.size:
        # If image_inds is given match image_inds to the data sets
        tile_refs = []
        for im_nr in image_inds:
            logger.debug(
                "Trying the find image nr {} in hdf5 file".format(
                    im_nr))
            try:
                found_tile = hdf5_group[str(im_nr)]
                cur_ref = group_root_name + str(im_nr)
            except KeyError:
                cur_ref = None
            logger.debug("Appending reference: {}".format(cur_ref))
            tile_refs.append(cur_ref)
        return tile_refs
    else:
        # If no image_inds are given, return all data set references
        # in alphabetical order (relies on hdf5 alphabetical return
        # of keys)
        tile_refs = [group_root_name + tile_name for tile_name in
                     hdf5_group.keys()]
        logger.debug("Got all refs: {}".format(tile_refs))
        return tile_refs


def get_image_names(hdf5_file, image_list=np.array([]),
                    hyb_nr=1, gene='Nuclei', pre_proc_level='FilteredData'):
    """Get the references to the tiles from the hdf5 file.

    If no matching data set was found, None will be used
    instead of a string reference.
    
    Parameters:
    -----------

    hdf5_file: pointer
        HDF5 file handle. The file containing the tiles.
    image_list: iterable
        Iterable containing ints. Should contain the
        names of the data sets that will be selected. This
        list could also contain strings, as long as it's
        contents reflect a subset of dataset names.
    hyb_nr: int
        The number of the hybridization where the
        data sets can be found. This will be used to
        navigate tile_file and find the correct tiles.
        (Default: 1)
    gene: str
        The name of the gene where the data sets can be found.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Nuclei')
    pre_proc_level: str
        The name of the pre processing group
        where the data sets can be found.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Filtered')
    """
    #logger = logging.getLogger(__name__)
    #logger.info("Getting image references from hdf5 file...")
    # Get the right group
    cur_group = hdf5_file[gene][pre_proc_level]
    return extract_hdf5_refs(cur_group, image_list)


def load_tile(file_name, hdf5_file=None):
    """Load a tile into memory as 2D image.

    Parameters:
    -----------

    file_name: str 
        Hdf5 reference to tile data set.
    hdf5_file: pointer
        HDF5 file handle. Opened HDF5 file containing the
        tile.
    
    Returns:
    --------

    : np.array 
            The tile as 2D image. Or an empty array when the file_name is None.
    """
    #logger = logging.getLogger(__name__)
    #logger.info("Loading image from: {}".format(file_name))
    if file_name is not None:
        # Flatten the image
        # Flattening also causes the image to be copied from the
        # hdf5-file into working memory as numpy array. It is
        # important to have the image as a numpy array rather than a
        # hdf5 dataset, because we want to do fancy slicing, which
        # is not possible on a dataset.
        return np.amax(hdf5_file[file_name], axis=0)
    else:
        return np.array([])


def load_tile_3D(file_name, hdf5_file=None):
    """Load a tile into memory as 3D image.

    Parameters:
    -----------

    file_name: str
        Hdf5 reference to tile data set.
    hdf5_file: pointer
        HDF5 file handle. Opened HDF5 file containing the tile.
    
    Returns:
    --------

    : np.array
        The tile as 3D image. Or an empty array when the file_name is None.
    """
    logger = logging.getLogger(__name__)
    logger.info("Loading image from: {}".format(file_name))

    if file_name is not None:
            # Copy the image from the hdf5-file into working memory as
            # numpy array. It is important to have the image as a numpy
            # array rather than a hdf5 dataset, because we want to do
            # fancy slicing, which is not possible on a dataset.
            return hdf5_file[file_name][()]
    else:
        return np.array([])


def save_image(im_file, hyb_nr, gene, pre_proc_level, image_name,
               save_file_name):
    """
    Save hdf5 file as a tiff image

    Save the data found in im_file_name[hyb_nr][gene][pre_proc_level][
    image_name] as an image in a tiff-file called save_file_name.
    im_file should be an hdf5-file and image_name should be a
    variable in that file.
    
    Parameters:
    -----------

    im_file: pointer
        HDF5 file handle. Opened HDF5 file containing the image to be saved.
    hyb_nr: int 
        Number of the current hybridization.
    gene: str 
        The name of the gene where the
        image to be saved can be found.
        This will be used to navigate im_file and find the
        correct image.
    pre_proc_level: str
        The name of the pre processing group
        where the image to be saved can be found.
        This will be used to navigate im_file and find the
        correct image.
    image_name: str
        The name of the dataset containing the image to be saved.
    save_file_name: str
        The name of the image file that is going to be generated, should be given without
        extension.
    """
    logger = logging.getLogger(__name__)
    logger.info("Saving image as: {}.tif".format(save_file_name))
    io.use_plugin('tifffile')
    #image_group = im_file['Hybridization' + str(hyb_nr)][gene][
    #    pre_proc_level]
    image_group = im_file[gene][pre_proc_level]

    io.imsave(save_file_name + ".tif", image_group[image_name],
              plugin="tifffile")


def save_to_file(file_name, **kwargs):
    """Save pickled stitching info to file_name

    Parameters:
    -----------

    file_name: str
        The name of the file where the data will be saved. Should be given without extension.
    
    **kwarg
        All keyword argument values will be
        gathered in a dictionary, with the variable name
        as their key in the dictionary. This dictionary will
        then be saved in the pickled file.

    """
    # Add extension
    file_name = file_name + '.pkl'
    logger = logging.getLogger(__name__)
    logger.info("Saving data to: {}".format(file_name))
    # Make a dictionary
    data = {}
    for key, val in kwargs.items():
        data[key] = val
    logger.debug("Saving data: {}".format(data))
    # Write to file
    with open(file_name, 'wb') as f:
        pickle.dump(data, f)


def load_stitching_coord(file_name):
    """Load pickled variables

    Parameters:
    -----------

    file_name: str
        The name of a pickled data file. Should be given without extension.
    
    Returns:
    --------

    : dict  
        Unpickled data found in <file_name>'.pkl'
    """
    file_name = file_name + '.pkl'
    logger = logging.getLogger(__name__)
    logger.debug("Loading data from: {}".format(file_name))
    with open(file_name, 'rb') as f:
        return pickle.load(f)


def free_hdf5_space(f):
    """Clean up the hdf5 file after deleting a dataset

    After deleting a data set from an hdf5-file, the disk space taken
    up by that data is not freed. Flushing the hfd5-file may help to
    keep the file smaller after deleting several data sets.

    Parameters:
    -----------

    f: pointer
        HDF5 file handle. The opened file that needs flushing.

    """
    logger = logging.getLogger(__name__)
    logger.info(
        "Flushing hdf5 file to clean up after delete operations")
    before_flush = f.id.get_filesize()
    f.flush()
    after_flush = f.id.get_filesize()
    logger.debug(
        "Size in bytes before flush: {} after flush: {} space freed: {}".format(
            before_flush, after_flush, before_flush - after_flush))


def read_IJ_corners(folder):
    """Read Preibisch corners

    Read previously determined global corners from the Preibisch
    imageJ plugin data. This code starts reading below the line:
    '# Define the image coordinates' and the file should not contain
    empty lines after this line.

    Parameters:
    -----------

    folder: str  Exact path to the folder, including trailing "/"

    Returns:
    --------

    compare_corners: list
        List, each element is a list containing the
        tile_ind and a numpy arrray with the
        corresponding corner's y and x coordinates.

    """
    logger = logging.getLogger(__name__)
    # Key word to look for in the name of the  file with the previously calculated corners
    name_key = 'TileConfiguration'
    # Find the file, using the key word
    coord_file_name = next((name for name in glob.glob(folder + '*.txt')
                            if name_key in name), None)
    logger.info(
        "Reading corners from folder: {} and file: {}".format(folder,
                                                              coord_file_name))
    with open(coord_file_name, 'r') as coord_file:
        coord_reader = csv.reader(coord_file, delimiter=';')
        start_reading = False
        # Init corner list
        compare_corners = []
        for line in coord_reader:
            logger.debug(line)
            if start_reading:
                # Take the last number found in the string as number of the tile
                tile_nr = int(re.findall('\d+', line[0])[-1])
                # Convert corner coordinates
                corner = ast.literal_eval(line[-1].strip())
                # Append the data we want to use:
                compare_corners.append([tile_nr, corner])
                logger.debug('{} {}'.format(tile_nr, corner))
            # Read from '# Define the image coordinates' till the end
            if line == ['# Define the image coordinates']:
                # Found starting line
                logger.debug('starting to read')
                start_reading = True
    return compare_corners


######################## Plotting ######################################
def display_tiles(tiles, tile_set, fig_nr=1, block=True,
                  maximize=False, main_title=None, rgb=False):
    """Plot tiles in subplots

    Parameters:
    -----------

    tiles: list
        List of tiles
    tile_set: np.array
        Tiles organization
    fig_nr: int     
        number of the figure window.
    block: bool
        Block causes the program to stop running after plotting the figure,
        until the figure window is closed.
    maximize: bool 
        Maximize plots the figure full screen.
    main_title: str
        Main title adds a title text in the figure.
    rgb: bool
        define the color standard used
    """
    if plt_available:
        logger = logging.getLogger(__name__)
        logger.info("Plotting tiles")
        plt.figure(fig_nr)
        for i in range(len(tiles)):
            plt.subplot(tile_set.shape[0], tile_set.shape[1], i + 1)
            if tiles[i].ndim > 2 and not (rgb):
                plt.imshow(tiles[i][0, :, :], 'gray',
                           interpolation='none')
            else:
                plt.imshow(tiles[i], 'gray', interpolation='none')
        if maximize:
            logger.debug(
                "Matplotlib backend: {}".format(plt.get_backend()))
            figManager = plt.get_current_fig_manager()
            figManager.window.showMaximized()
        if main_title:
            plt.suptitle(main_title, fontsize=20)
        plt.show(block=block)
    else:
        return None


def plot_stitched_coordinates(corner_list, image_list, block=True,
                              name='stitched_coordinates',
                              file_name=None,
                              invert_yaxis=True):
    """Plot top-left coordinates with their tile numbers

    Coordinates are assumed to be in pixels.
    
    Parameters:
    -----------
    corner_list : list
        List with as many elements as there are tiles. Each element contains
        the tile index (as int) and y and x coordinate of the top left corner
        (as a numpy array)
    image_list: list
        The list of the numbers of the tiles as they are found in the file names,
        but reordered to match the order of corner_list.
    block: bool
        Plot blocks the running program untill the plotting window is closed if true
        (default: True)
    name: str 
        Name of the plotting window (default: "coordinates")
    file_name: str
        Name of the file the image should be saved to.
        The name should list the complete absolute map.
        Default/None causes the image to not be saved. (Default: None)
    invert_yaxis: bool
        Invert the y-axis when plotting. (default: True)
    """
    if plt_available:
        labels = [str(corner_list[i][0]) + '/' + str(image_list[i])
                  for i in range(len(corner_list))]
        corners = np.array([corner[1] for corner in corner_list])
        fig = plt.figure(name)
        plt.plot(corners[:, 1], corners[:, 0], 'o')
        # plt.setp(plt.gca(), 'ylim', reversed(plt.getp(plt.gca(), 'ylim')))
        # bboxes = []
        # ax = plt.gca()
        for label, x, y in zip(labels, corners[:, 1], corners[:, 0]):
            annotation = plt.annotate(
                label,
                xy=(x, y), xytext=(-2, 2),
                textcoords='offset points', ha='center', va='bottom')
            # plt.draw()
            # fig.canvas.draw()
            # bbox = annotation.get_window_extent()
            # the figure transform goes from relative coords->pixels and we
            # want the inverse of that
            # bbox_data = bbox.transformed(ax.transData.inverted())
            # bboxes.append(bbox_data)
        # Title and axes labels
        plt.title(
            'Labels represent: <tile index>/<image number as it appears in file name>',
            y=1.08)
        plt.xlabel('top-left position (pxls)')
        plt.ylabel('top-left position (pxls)')

        ymin, ymax = plt.ylim()  # return the current ylim
        plt.ylim(
            (ymin - 1000, ymax + 1000))  # set the ylim to ymin, ymax
        xmin, xmax = plt.xlim()  # return the current ylim
        plt.xlim(
            (xmin - 1000, xmax + 1000))  # set the ylim to ymin, ymax

        # bbox = mtransforms.Bbox.union(bboxes)
        # ax.update_datalim(bbox.corners())
        # ax.autoscale_view()

        if invert_yaxis:
            ax = plt.gca()
            plt.setp(ax, 'ylim', reversed(plt.getp(plt.gca(), 'ylim')))
            ax.xaxis.tick_top()
            ax.xaxis.set_label_position('top')
        plt.tight_layout()
        # Make sure the labels are within the axes
        # plt.draw()
        # bboxes = []
        # for bbox in bboxes:
        # bbox = annotation.get_window_extent()
        # the figure transform goes from relative coords->pixels and we
        # want the inverse of that
        # bbox_data = bbox.transformed(ax.transData.inverted())
        # bboxes.append(bbox_data)



        # Save and/or display
        if file_name is not None:
            plt.savefig(file_name + '.png')
        plt.show(block=block)
    else:
        return None


def plot_coordinates(micData, ind_coord_list=None, block=True,
                     name="coordinates",
                     file_name=None, invert_yaxis=True):
    """Plot top-left coordinates with their tile numbers

    Coordinates are assumed to be in pixels.
    
    Parameters:
    -----------
    
    micData: object
        MicroscopeData object. Containing the coordinates to be plotted.
    block: bool
        Plot blocks the running program untill the plotting window is closed if true
        (default: True)
    name: str
        Name of the plotting window (default: "coordinates")
    file_name: str
        Name of the file the image should be saved to.
        The name should list the complete absolute map.
        Default/None causes the image to not be saved. (Default: None)
    invert_yaxis: bool
        Invert the y-axis when plotting. (default: True)
    """
    if plt_available:
        # Define labels
        if ind_coord_list is None:
            labels = [str(nr) for nr in micData.tile_nr]
        else:
            labels = [str(ind_coord_list[i][0]) + '/' + str(
                ind_coord_list[i][1])
                      for i in range(len(ind_coord_list))]
        plt.figure(name)
        plt.plot(micData.x_coords, micData.y_coords, 'o--')
        # Add annotations
        for label, x, y in zip(labels, micData.x_coords,
                               micData.y_coords):
            plt.annotate(
                label,
                xy=(x, y), xytext=(-2, 2),
                textcoords='offset points', ha='center', va='bottom')
        # Axes labels
        plt.xlabel('top-left position (pxls)')
        plt.ylabel('top-left position (pxls)')
        # Enlarge axes so the annotation will fit within the axis better
        ymin, ymax = plt.ylim()  # return the current ylim
        plt.ylim(
            (ymin - 1000, ymax + 1000))  # set the ylim to ymin, ymax
        xmin, xmax = plt.xlim()  # return the current ylim
        plt.xlim(
            (xmin - 1000, xmax + 1000))  # set the ylim to ymin, ymax

        if invert_yaxis:
            ax = plt.gca()
            plt.setp(ax, 'ylim', reversed(plt.getp(plt.gca(), 'ylim')))
            ax.xaxis.tick_top()
            ax.xaxis.set_label_position('top')

        # Save and/or display
        if file_name is not None:
            plt.savefig(file_name + '.png')
        plt.show(block=block)
    else:
        return None


def plot_3D(image):
    if plt_available:
        logger = logging.getLogger(__name__)
        logger.info("Plotting tiles in 3D")
        logger.debug("Shape pic: {}".format(image.shape))
        plt.figure('test final img')
        plt.imshow(image[0, :, :])
        plt.show()

        np_image = np.array(image)
        app = vv.use()
        vv.volshow(np_image, renderStyle='mip')
        a = vv.gca()
        a.camera.fov = 45
        a.daspectAuto = False
        a.daspect = (1, 1, 10)
        app.Run()
    else:
        return None



from sympy import Point, Line, Segment
from skimage import feature, measure
import numpy as np
from scipy import ndimage as nd

def thr_calculator(filtered_img,min_distance=3,stringency=0):

    """
    Function used to calculate the threshold to use for the dots
    counting in a 2D image. 

    Parameters:
    -----------

    filtered_img: np.array float64
        preprocessed image used to count the dots.
    min_distance: int
        minimum distance that two maxima need to have in order to be defined as 
        separete peaks.
    stringency: int
        integer used to select the stringency of the generated
        threshold. By adding stringency to the thr_idx we can select a Thr with higher
        value from the thr_array.

    Returns:
    -----------

    counting_dict : dict 
        dictionary containing all the counting infos:
        selected_thr: float64
            Thr used for counting after application of the stringency.
        calculated_thr: float64 
            Calculated Thr
        selected_peaks: int64 
            2D coords of the peaks defined using the selected_thr.
        thr_array: float64 
            Thr array of 100 points distributed between (Img.min(),Img.max()).
        peaks_coords: float64 
            list of all the 3D coords calculated using the Thr array.
        total_peaks: list of int 
            List of the peaks counts.
        thr_idx: int64 
            index of the calculated threshold.
        stringency: int64 
            stringency used for the identification of the selected_peaks
    """
    
    # List with the total peaks calculated for each threshold
    total_peaks = []
    
    # List of ndarrays with the coords of the peaks calculated for each threshold
    peaks_coords = []

    # Define the Thr array to be tested
    thr_array = np.linspace(filtered_img.min(),filtered_img.max(),num=100)


    # Calculate the number of peaks for each threshold. In this calculation
    # the size of the objects is not considered
    for thr in thr_array:
        # The border is excluded from the counting
        peaks = feature.peak_local_max(filtered_img,min_distance=min_distance,\
            threshold_abs=thr,exclude_border=False, indices=True,\
            num_peaks=np.inf, footprint=None,labels=None)    
        # Stop the counting when the number of peaks detected falls below 3
        if len(peaks)<=3:
            stop_thr = thr # Move in the upper loop so you will stop at the previous thr
            break
        else:
            peaks_coords.append(peaks) 
            total_peaks.append(len(peaks))


    # Consider the case of no detectected peaks or if there is only one Thr
    # that create peaks (list total_peaks have only one element and )
    # if np.array(total_peaks).sum()>0 or len(total_peaks)>1:
    if len(total_peaks)>1:

        # Trim the threshold array in order to match the stopping point
        # used the [0][0] to get the first number and then take it out from list
        thr_array = thr_array[:np.where(thr_array==stop_thr)[0][0]]


        # Calculate the gradient of the number of peaks distribution
        grad = np.gradient(total_peaks)
        
        # Restructure the data in order to avoid to consider the min_peak in the
        # calculations

        # Coord of the gradient min_peak
        grad_min_peak_coord = np.argmin(grad)
        
        # Trim the data to remove the peak.
        trimmed_thr_array = thr_array[grad_min_peak_coord:]
        trimmed_grad = grad[grad_min_peak_coord:]

        if trimmed_thr_array.shape>(1,):

            # Trim the coords array in order to maintain the same length of the 
            # tr and pk
            trimmed_peaks_coords = peaks_coords[grad_min_peak_coord:]
            trimmed_total_peaks = total_peaks[grad_min_peak_coord:]

            # To determine the threshold we will determine the Thr with the biggest
            # distance to the segment that join the end points of the calculated
            # gradient

            # Distances list
            distances = []

            # Calculate the coords of the end points of the gradient
            p1 = Point(trimmed_thr_array[0],trimmed_grad[0])
            p2 = Point(trimmed_thr_array[-1],trimmed_grad[-1])
            
            # Create a line that join the points
            s = Line(p1,p2)
            allpoints = np.arange(0,len(trimmed_thr_array))
            
            # Calculate the distance between all points and the line
            for p in allpoints:
                dst = s.distance(Point(trimmed_thr_array[p],trimmed_grad[p]))
                distances.append(dst.evalf())

            # Remove the end points from the lists
            trimmed_thr_array = trimmed_thr_array[1:-1]
            trimmed_grad = trimmed_grad[1:-1]
            trimmed_peaks_coords = trimmed_peaks_coords[1:-1]
            trimmed_total_peaks = trimmed_total_peaks[1:-1]
            trimmed_distances = distances[1:-1]
        
            # Determine the coords of the selected Thr
            # Converted trimmed_distances to array because it crashed
            # on Sanger.
            if trimmed_distances: # Most efficient way will be to consider the length of Thr list
                thr_idx=np.argmax(np.array(trimmed_distances))
                calculated_thr = trimmed_thr_array[thr_idx]
                # The selected threshold usually causes oversampling of the number of dots
                # I added a stringency parameter (int n) to use to select the Thr+n 
                # for the counting. It selects a stringency only if the trimmed_thr_array
                # is long enough
                if thr_idx+stringency<len(trimmed_thr_array):
                    selected_thr = trimmed_thr_array[thr_idx+stringency]
                    selected_peaks = trimmed_peaks_coords[thr_idx+stringency]
                else:
                    selected_thr = trimmed_thr_array[thr_idx]
                    selected_peaks = trimmed_peaks_coords[thr_idx]

                
                # Calculate the selected peaks after removal of the big and small objects
                
                # Threshold the image using the selected threshold
                if selected_thr>0:
                    img_mask = filtered_img>selected_thr
                
                labels = nd.label(img_mask)[0]
                
                properties = measure.regionprops(labels)
                    
                for ob in properties:
                    if ob.area<6 or ob.area>200:
                        img_mask[ob.coords[:,0],ob.coords[:,1]]=0
                
                labels = nd.label(img_mask)[0]
                selected_peaks = feature.peak_local_max(filtered_img, min_distance=min_distance, threshold_abs=selected_thr, exclude_border=False, indices=True, num_peaks=np.inf, footprint=None, labels=labels)
                
                if selected_peaks.size:
                    # Intensity counting of the max peaks
                    selected_peaks_int = filtered_img[selected_peaks[:,0],selected_peaks[:,1]]
                
                
                else:
                    selected_thr = 0
                    calculated_thr = 0
                    selected_peaks = 0
                    peaks_coords = 0
                    total_peaks = 0
                    thr_idx = 0
                    selected_peaks_int = 0
                           
            else:
                selected_thr = 0
                calculated_thr = 0
                selected_peaks = 0
                peaks_coords = 0
                total_peaks = 0
                thr_idx = 0
                selected_peaks_int = 0
        else:
            selected_thr = 0
            calculated_thr = 0
            selected_peaks = 0
            peaks_coords = 0
            total_peaks = 0
            thr_idx = 0
            selected_peaks_int = 0


    else:
        selected_thr = 0
        calculated_thr = 0
        selected_peaks = 0
        peaks_coords = 0
        total_peaks = 0
        thr_idx = 0
        selected_peaks_int = 0

    counting_dict={}

    counting_dict['selected_thr'] = selected_thr
    counting_dict['calculated_thr'] = calculated_thr
    counting_dict['selected_peaks'] = selected_peaks
    counting_dict['thr_array'] = thr_array
    counting_dict['peaks_coords'] = peaks_coords
    counting_dict['total_peaks'] = total_peaks 
    counting_dict['thr_idx'] = thr_idx
    counting_dict['stringency'] = stringency
    counting_dict['selected_peaks_int'] = selected_peaks_int
    
    return counting_dict



import numpy as np
import sklearn.linear_model as linmod
import logging
# import matplotlib.pyplot as plt

class GlobalOptimization:
    """
    Use linear regression to find the global transitions that fit
    the pairwise transitions best

    Variables:
    ----------
    self.logger         -- logger instance
    self.global_trans   -- 2D numpy array containg the global
                        translation for each tile. (Or empty list before
                        running performOptimization())
    """


    def __init__(self):
        """Initialize logger and an array to save global transitions."""
        self.logger = logging.getLogger(__name__)
        self.global_trans = []              # array, shape: (nr_tiles, nr_dim)

    def performOptimization(self, tile_set, contig_tuples, P, covs,
                            nr_tiles, nr_dim):
        """Use linear regression to find the global transition

        Fills up self.global_trans; a numpy array with shape (nr_tiles,
        nr_dim).

        Parameters:
        -----------
        tile_set: np.array
            Array filled with ones that has the same shape as tile set
        contig_tuples: list
            List of tuples denoting which tiles are contingent to each other.
        P: np.array
            1D numpy array containing pairwise alignment y and x coordinates 
            (and z-coordinates when applicable) for each neighbouring pair of
            tiles, array should be 2 * len(contig_typles) f for 2D data or 
            3 * len(contig_typles) for 3D data.
        covs: np.array
            Covariance for each pairwise alignment in P, array should be 
            len(contig_typles).
        nr_tiles: int
            The number of tiles in the dataset
        nr_dim: int 
            The number of dimensions the image.
        """
        self.logger.info("Calculating global transitions...")

        Z_xlen = tile_set.shape[0] * tile_set.shape[1] * nr_dim

        # Prepare covariance to function as weight
        weights = np.zeros(len(P))

        # Build Design matrix, P = ZQ * Unknown
        # Order of building: Y, X, Z
        ZQ = np.zeros((len(contig_tuples) * nr_dim, Z_xlen))
        for i, (a, b) in enumerate(contig_tuples):
            # Y row:
            Z = np.zeros((Z_xlen))
            Z[nr_dim * a:nr_dim * a + 1] = -1
            Z[nr_dim * b:nr_dim * b + 1] = 1
            ZQ[i * nr_dim, :] = Z
            # X row
            Z = np.zeros((Z_xlen))
            Z[nr_dim * a + 1:nr_dim * a + 2] = -1
            Z[nr_dim * b + 1:nr_dim * b + 2] = 1
            ZQ[i * nr_dim + 1, :] = Z
            # Z row:
            if nr_dim == 3:
                Z = np.zeros((Z_xlen))
                Z[nr_dim*a + 2:nr_dim * a + 3] = -1
                Z[nr_dim*b + 2:nr_dim * b + 3] = 1
                ZQ[i * nr_dim + 2,:] = Z
            # Prepare covariance to function as weight
            if np.isnan(covs[i]):
                weights[i * nr_dim:i * nr_dim + nr_dim] = 0.0
            else:
                weights[i * nr_dim:i * nr_dim + nr_dim] = covs[i]
        self.logger.debug("weights: {}".format(weights))
        #self.logger.debug("ZQ: {}".format(ZQ))
        # plt.figure()
        # plt.imshow(ZQ, interpolation='none')
        # plt.show()
        # Find unknown in P = ZQ * Unknown through linear regression
        lrg = linmod.LinearRegression(fit_intercept=False)
        lrg.fit(ZQ, P)
        #lrg.fit(ZQ, P, weights)

        # Get coordinates with respect to zero point (tile (0,0))
        self.logger.debug("Nr of tiles: {}, nr of results lrg: {}"
                            .format(nr_tiles, len(lrg.coef_)))
        self.global_trans = lrg.coef_.reshape((nr_tiles, nr_dim))
        self.global_trans = -1 * (-lrg.coef_.reshape((nr_tiles, nr_dim)) \
                            + lrg.coef_.reshape((nr_tiles, nr_dim))[0:1, :])
        self.logger.info("Global transition calculated, global trans:\n{}"
                            .format(self.global_trans))




# Check the imports that are not required

# General imports
import argparse
from mpi4py import MPI
import glob
import h5py
from skimage import filters,io,img_as_float,exposure,morphology,segmentation,measure,feature,color,img_as_ubyte
import os
from scipy import ndimage as nd
import glob
import re
import joblib
import numpy as np
import itertools
import shutil

# Own imports
from pysmFISH import utils
from pysmFISH import MPI_utils
from pysmFISH import dots_calling
from pysmFISH import h5writingfun

# ---------------- Functions --------------------------

def crop_object_coords(data_file_hdl,obj,segmentation_gene):
    
    # Coords of the obj
    row_Coords_obj=data_file_hdl[analysis_name]['Segmentation'][segmentation_gene]['Objects'][obj][:,0]
    col_Coords_obj=data_file_hdl[analysis_name]['Segmentation'][segmentation_gene]['Objects'][obj][:,1]
    
    # Calculate the coords of the bounding box to use to crop the ROI
    # Added +1 because in the cropping python will not consider the end of the interval
    bb_row_max=row_Coords_obj.max()+1
    bb_row_min=row_Coords_obj.min()
    bb_col_max=col_Coords_obj.max()+1
    bb_col_min=col_Coords_obj.min()
    
    # Normalize the coords of the obj
    row_Coords_obj_norm=row_Coords_obj-bb_row_min
    col_Coords_obj_norm=col_Coords_obj-bb_col_min
    
    
    return (bb_row_min,bb_row_max,bb_col_min,bb_col_max),(row_Coords_obj_norm,col_Coords_obj_norm)

# Check on how you can merge dictionary
# datatype non used but necessary in order to avoid a data type error in the 
# implementation of MPI
def addCounter(counter1, counter2, datatype):
    for item in counter2:
            counter1[item] = counter2[item]
    return counter1


# ---------------- MPI parameters --------------------------
comm = MPI.COMM_WORLD
size = comm.size        # total number of processes
rank = comm.rank        # rank of this process


# # input
parser = argparse.ArgumentParser(description='Script counting dots')
parser.add_argument('-d', help='Experimental directory')
parser.add_argument('-n',help='Hybridization Number')
parser.add_argument('-f',help='Preprocessing level (ex. StitchedImageRegistered)')
parser.add_argument('-x',help='AnalysisName')
parser.add_argument('-g',help='segmentation gene')


args= parser.parse_args()
data_folder=args.d
hyb_nr=int(args.n)
FolderLevel=args.f
analysis_name=args.x
segmentation_gene=args.g
# inputs for testing
# analysis_name='AnalysisNovember2016'
# hyb_nr=1
# segmentation_gene='polyA'
# Folder containing the image to use for counting
# FolderLevel='StitchedImageRegistered'
# data_folder='/data/users/simonec/Data_Simone/smFISH/Cortex/Stitched/'

skipping=['Nuclei','polyA']



# ------------- File Loading --------------------------------
# # Open the the file with the hybridization to count
stitched_file_name = glob.glob(data_folder + '*_Hybridization' +
                                str(hyb_nr) + '.sf.hdf5')[0]

# # for testing
# stitched_file_name='/data/users/simonec/Data_Simone/smFISH/Cortex/Stitched/AnalysisNovember2016_Hybridization1.sf.hdf5'

stitched_file_hdl=h5py.File(stitched_file_name,'r',driver='mpio',comm=MPI.COMM_WORLD)

# ------------- File Analysis --------------------------------


# data_file_name='/data/users/simonec/Data_Simone/smFISH/Cortex/Stitched/AnalysisNovember2016Analysis.data.hdf5'

data_file_name=data_folder+'Analysis.data.hdf5'
# Open in rank 0 than bcast the coords and bb_coords to the other cores


data_file_hdl=h5py.File(data_file_name,'r+',driver='mpio',comm=MPI.COMM_WORLD)

# Create list of genes to process
ToKeep=set(stitched_file_hdl)-set(skipping)
genes_list=[gene for gene in ToKeep if '_IF' not in gene]

# # Create the gene groups
# if genes_list:
#     for gene in stitched_file_hdl:
#         if gene not in ['Nuclei','polyA'] or '_IF' not in gene:
#             gene_grp=data_file_hdl[analysis_name].create_group(gene)


# if genes_list:

# IT IS WORTH TO WRITE IN PARALLEL TO SPEED UP THE PROCESS

# Try to avoid parallel writing on hdf5 so you can make it easier to move to
# a non MPI and more pythonic system

# Presave the data to avoid to used parallel hdf5....also difficult because I
# don't know the size of the datasets than need to be pre-created on all cores
# I will have to create a resizable dataset
# Create temporary folder for saving the counting
if rank==0:

    for gene in genes_list:
            DirPath=data_folder+'RNA_counting_tmp_Hybridization'+str(hyb_nr)+'_'+gene
            try:
                os.stat(DirPath)
            except:
                os.mkdir(DirPath)
                os.chmod(DirPath,0o777)


# ------------- Scatter the obj regions for counting ---------------------------
if rank==0:
    
    Objects=list(data_file_hdl[analysis_name]['Segmentation'][segmentation_gene]['Objects'].keys())

    TasksList=np.array([np.int32(obj) for obj in Objects])
     # Chunk the distribution list
    Chunked_list,NumberChunks,Displacement = MPI_utils.tasks_chunking(TasksList,size)


else:
    NumberChunks=None
    Displacement=None
    Chunked_list=None
    TasksList=None

TasksList=comm.bcast(TasksList,root=0)
# Make the cores aware of the number of jobs that will need to run
# The variable count is created by the scatter function and has the number of
# processes and is different in every core
cnt = np.zeros(1, dtype='int32')
comm.Scatter(NumberChunks, cnt, root=0)
# Define the local variable that will be filled up wuth the scattered data
xlocal = np.zeros(cnt, dtype='int32') # Use rank for determine the size of the xlocal on the different cores
# Scatter the value of tasks to the different cores
comm.Scatterv([TasksList,NumberChunks,Displacement,MPI.INT],xlocal, root=0)

# Create a coords dict on each rank
# It may not be efficient to calculate it every time an hyb is processed. I save
# in memory so will be used for all the channels in the hyb. I think that saving it
# and load it may be slower
Coords_dict=dict()
for obj in xlocal:
    bb_coords,obj_coords=crop_object_coords(data_file_hdl,str(obj),segmentation_gene)
    Coords_dict[obj]={'bb_coords':bb_coords,'obj_coords':obj_coords}

# process each gene separetely. I could also create a massive list but as 
# starting point I will process each gene
for gene in genes_list:
    tmp_folder_path=data_folder+'RNA_counting_tmp_Hybridization'+str(hyb_nr)+'_'+gene+'/'
    for obj in xlocal:
        bb_coords=Coords_dict[obj]['bb_coords']
        obj_coords=Coords_dict[obj]['obj_coords']
        
        # Load the cropped image
        Img_cropped=stitched_file_hdl[gene][FolderLevel]['final_image'][bb_coords[0]:bb_coords[1],bb_coords[2]:bb_coords[3]]
        
        # Mask the image according to the segmented region
        Img_masked=np.zeros_like(Img_cropped)
        Img_masked[obj_coords[0],obj_coords[1]]=Img_cropped[obj_coords[0],obj_coords[1]]

        Selected_Thr,Calculated_Thr,Selected_Peaks,ThrArray,PeaksCoords,\
             TotalPeaks,Thr_idx,stringency,Selected_peaks_int=dots_calling.ThrCalculator(Img_masked,min_distance=5,stringency=0,dimensions=2)
        
        
#         # Consider black image case
        if Selected_Thr>0:
#             # Intensity counting of the max peaks
#             Selected_peaks_int=Img_masked[Selected_Peaks[:,0],Selected_Peaks[:,1]]

            # Dots coords correction (in order to be able to map the dots on the)
            # stitched image and not only in the selected region.
            
            # Recalculate the coords of all the peaks (with different thresholds)
            PeaksCoords_reMapped_tmp=[]
            PeaksCoords_reMapped=[]
            for peaks_coords in PeaksCoords:
                for peak in peaks_coords:
                    NewCoords=[peak[0]+bb_coords[0],peak[1]+bb_coords[2]]
                    # NewCoords=peak+[bb_coords[0],bb_coords[2]]
                    PeaksCoords_reMapped_tmp.append(NewCoords)
                PeaksCoords_reMapped.append(PeaksCoords_reMapped_tmp)
                PeaksCoords_reMapped_tmp.append(NewCoords)
            
            # Recalculate the coords of the selected peaks
            Selected_Peaks_reMapped=[]
            for peak in Selected_Peaks:
                NewCoords=[peak[0]+bb_coords[0],peak[1]+bb_coords[2]]
                # NewCoords=peak+[bb_coords[0],bb_coords[2]]
                Selected_Peaks_reMapped.append(NewCoords)
        else:
            PeaksCoords_reMapped_tmp=PeaksCoords
            Selected_Peaks_reMapped=Selected_Peaks

        # Save the tmp file
        # Save the data in a compressed format before loading them into a hdf5_file
        joblib.dump((Selected_Thr,Calculated_Thr,Selected_Peaks_reMapped,\
            Selected_peaks_int,ThrArray,PeaksCoords_reMapped_tmp,\
            TotalPeaks,Thr_idx,stringency,bb_coords),tmp_folder_path+'Hybridization'+str(hyb_nr)+'_'+gene+'_PeaksData_obj_'+str(obj)+'.pkl')


stitched_file_hdl.close()
data_file_hdl.close()

# # Write everything on the Analysis Results file
if rank==0:
    data_file_hdl=h5py.File(data_file_name,'r+',swmr=True,libver='latest')
    SegmentedRegionsCounts_grp=data_file_hdl[analysis_name].require_group('SegmentedRegionsCounts')
    HybRef='Hybridization'+str(hyb_nr)
    Hybridization_grp=SegmentedRegionsCounts_grp.require_group(HybRef)
    for gene in genes_list:
        _=Hybridization_grp.require_group(gene)
        for obj in Objects:   
            tmp_folder_path=data_folder+'RNA_counting_tmp_Hybridization'+str(hyb_nr)+'_'+gene+'/'  
            h5writingfun.serial_writing_dots_segreg(data_file_hdl,analysis_name,tmp_folder_path,\
                hyb_nr,gene,obj)
        # delete the tmp folders
        shutil.rmtree(tmp_folder_path)

    data_file_hdl.close()




"""
Functions that handle hdf5 files
"""

import h5py
import numpy as np
import glob
from collections import OrderedDict
import pickle
import loompy


def hdf5_create_preprocessing_file(hybridizations_infos,processing_hyb,hybridization,
                                   analysis_name, hyb_dir,converted_positions,
                                   image_properties):
    
    """ 

    Create the hdf5 file that will contain the preprocessed images
    for one hybridization.
    The preprocessing .hdf5 file must have a .ppf.hdf5 extension
    The file will be created for serial writing of the hdf5 file.
    After creation the file is closed.
    Remember that the 'lzf' compression do not accept extra items
    
    Parameters:
    -----------

    hybridizations_infos: dict 
        Parsed Hybridizations metadata
    processing_hyb: str
        Hybridization to process. The function will create an hdf5 file for the 
        hybridization only
    hybridization: str 
        Hybridization name
    analysis_name: str 
        Analysis name
    hyb_dir: str 
        Path of the dir where to save the file
    converted_positions: dict 
        Parsed positions
    """
    
    compression = 'lzf'
    preprocessing_file_name= analysis_name+'_'+processing_hyb+'.ppf.hdf5'
    preprocessing_file = h5py.File(hyb_dir+preprocessing_file_name, 'w',libver='latest')
    
    # ImagesInfo
    Hyb_cols_num_px=image_properties['HybImageSize']['columns']
    Hyb_rows_num_px=image_properties['HybImageSize']['rows']

    Strip_cols_num_px=image_properties['StripImageSize']['columns']
    Strip_rows_num_px=image_properties['StripImageSize']['rows']

    # Create the groups and the subgroups

    # List of the XY positions
    Positions_list=list(converted_positions[hybridization].keys())

    for gene in hybridizations_infos[hybridization].keys():
            gen_grp=preprocessing_file.create_group(gene)
            Fd_group=gen_grp.create_group('FilteredData')
            preprocessing_file.flush()
            if 'Hyb' in hybridization:

                for pos in Positions_list:
                    Fd_group.create_dataset(str(pos),dtype = np.uint16, \
                        shape=(Hyb_rows_num_px,Hyb_cols_num_px),\
                        compression=compression,\
                        shuffle=True,chunks=(Hyb_rows_num_px,Hyb_cols_num_px))
            else:
                for pos in Positions_list:
                    Fd_group.create_dataset(str(pos),dtype = np.uint16,\
                        shape=(Strip_rows_num_px,Strip_cols_num_px),\
                        compression=compression,\
                        shuffle=True,chunks=(Hyb_rows_num_px,Hyb_cols_num_px))

    preprocessing_file.close()
    preprocessing_file_path=hyb_dir + preprocessing_file_name
    
    return preprocessing_file_path




def create_loompy_raw_counting(hybridizations_infos,converted_positions,
                        hybridization,flt_rawcnt_config,hyb_dir,
                        processing_hyb,counting_gene_dirs):
    """
    Function used to write the counting results in a loom file

    Parameters:
    -----------
    
    hybridizations_infos: dict 
        Parsed information on the experiment.
    converted_positions: dict 
        coords of the images for all hybridization.
    hybridization: str 
        hybridization processed (ex. Hybridization2) 
    flt_rawcnt_config: dict
        Parsed filtering a raw counting configuration file 
        (Filtering_raw_counting.config.yaml)
    counting_gene_dirs: list 
        List of the directories containing the counts


    """

    # Loompy matrix column attributes
    pos_att = np.arange(0,len(converted_positions[hybridization].keys()))

    # Create dictionaries to store the data that have one array for each image
    # position

    counting_dict = OrderedDict()
    selected_peaks_dict = OrderedDict()
    thr_array_dict = OrderedDict()
    peaks_coords_dict = OrderedDict()
    total_peaks_dict = OrderedDict()
    selected_peaks_int_dict = OrderedDict()


    gene_idx_pos = 0
    gene_idx_list = []
    gene_list = list(hybridizations_infos[hybridization].keys())
    gene_list = [gene for gene in gene_list if gene not in flt_rawcnt_config['skip_genes_counting'] ]
    gene_list = [gene for tag in flt_rawcnt_config['skip_tags_counting'] for gene in gene_list if tag not in gene]

    # Create matrices for loompy layers
    total_counts_mat = np.zeros([len(gene_list),len(converted_positions[hybridization].keys())])
    selected_thr_mat = np.zeros([len(gene_list),len(converted_positions[hybridization].keys())])
    calculated_thr_mat = np.zeros([len(gene_list),len(converted_positions[hybridization].keys())])
    thr_idx_mat = np.zeros([len(gene_list),len(converted_positions[hybridization].keys())])
    stringency_mat = np.zeros([len(gene_list),len(converted_positions[hybridization].keys())])

    array_positions = np.sort(list(converted_positions[hybridization].keys()))

    if gene_list:
        for gene in gene_list:
            gene_idx_list.append(gene+'_'+hybridization)
            counting_gene_dir = [fdir for fdir in counting_gene_dirs if gene in fdir][0]
            counting_files_list = glob.glob(counting_gene_dir+'*.pkl')

            counting_dict[gene] =OrderedDict()

            counting_dict[gene]['selected_peaks'] =OrderedDict()
            counting_dict[gene]['thr_array'] =OrderedDict()
            counting_dict[gene]['peaks_coords'] =OrderedDict()
            counting_dict[gene]['total_peaks'] =OrderedDict()
            counting_dict[gene]['selected_peaks_int'] =OrderedDict()



            # Process the files according to sorted position
            for pos in array_positions:
                counting_file = [cf for cf in counting_files_list if 'pos_'+str(pos) in cf][0]
                countings = pickle.load(open(counting_file,'rb'))
                # pos = np.int(counting_file.split('/')[-1].split('_')[-1].split('.')[0])

                # loompy layers construction
                selected_thr_mat[gene_idx_pos,pos] = countings['selected_thr']
                calculated_thr_mat[gene_idx_pos,pos] = countings['calculated_thr']
                thr_idx_mat[gene_idx_pos,pos] = countings['thr_idx']
                stringency_mat[gene_idx_pos,pos] = countings['stringency']
                if isinstance(countings['selected_peaks'], list):
                    total_counts_mat[gene_idx_pos,pos] = len(countings['selected_peaks'])
                else:
                    total_counts_mat[gene_idx_pos,pos] = 0


                # Dictionaries construction

                counting_dict[gene]['selected_peaks'][pos] = countings['selected_peaks']
                counting_dict[gene]['thr_array'][pos] = countings['thr_array']
                counting_dict[gene]['peaks_coords'][pos] = countings['peaks_coords']
                counting_dict[gene]['total_peaks'][pos] = countings['total_peaks']
                counting_dict[gene]['selected_peaks_int'][pos] = countings['selected_peaks_int']



            gene_idx_pos +=1

    # Create loompy file
    loom_fname = hyb_dir+processing_hyb+'_raw_counting.loom'
    loom_hdl = loompy.create(filename=loom_fname,matrix=total_counts_mat,row_attrs={'genes':np.array(gene_idx_list)},col_attrs={'image_pos':np.array(pos_att)})
    loom_hdl.set_layer(name='selected_thr',matrix=selected_thr_mat)
    loom_hdl.set_layer(name='calculated_thr',matrix=calculated_thr_mat)
    loom_hdl.set_layer(name='thr_idx',matrix=thr_idx_mat)
    loom_hdl.set_layer(name='stringency',matrix=stringency_mat)
    loom_hdl.close()    

    # Add extra data to the loom file
    with h5py.File(loom_fname,'r+') as loom_hdl:
        counting_arrays_grp = loom_hdl.create_group('counting_arrays')
        for gene in counting_dict.keys():
            gene_grp = counting_arrays_grp.create_group(gene+'_'+hybridization)
            selected_peaks_grp = gene_grp.create_group('selected_peaks')
            thr_array_grp = gene_grp.create_group('thr_array')
            peaks_coords_grp = gene_grp.create_group('peaks_coords')
            total_peaks_grp = gene_grp.create_group('total_peaks')
            selected_peaks_int_grp = gene_grp.create_group('selected_peaks_int')
            for pos in array_positions:
                selected_peaks_grp.create_dataset(str(pos), data=counting_dict[gene]['selected_peaks'][pos])
                thr_array_grp.create_dataset(str(pos),data=counting_dict[gene]['thr_array'][pos])
                total_peaks_grp.create_dataset(str(pos),data=counting_dict[gene]['total_peaks'][pos])
                selected_peaks_int_grp.create_dataset(str(pos),data=counting_dict[gene]['selected_peaks_int'][pos])
                pos_sbgrp = peaks_coords_grp.create_group(str(pos))
                if isinstance(counting_dict[gene]['peaks_coords'][pos], list):
                    for idx,peaks in enumerate(counting_dict[gene]['peaks_coords'][pos]):
                        pos_sbgrp.create_dataset(str(idx),data=peaks)
                else:
                    pos_sbgrp.create_dataset(str(idx),data=peaks)


"""High level helper functions that deal with the alignment of
a single pair of neighbours, written for use by the class pairwise
alignment.
Creates its own logger object when imported.

Functions:
align_single_pair       -- Determine the ideal alignment between two
                        neighbouring tiles
refine_single_pair      -- Determine the ideal alignment between two
                        neighbouring tiles, with the use of an old
                        alignment.
determine_overlap       -- Determine the overlap between two
                        neighbouring tiles
calculate_pos_shifts    -- Calulate possible shifts, given two
                        overlapping images.
find_best_trans         -- Find the best translation using the cross
                        covariance.
find_best_trans_corr    -- Find the best translation using the cross
                        correlation.
perform_upsampling      -- Perform upsampling for subpixel precision in
                        the shift
"""

import numpy as np
from skimage import img_as_float
# import matplotlib.pyplot as plt
import logging

from . import inout
from . import pairwisehelper as ph


logger = logging.getLogger(__name__)

# The method used for the 3D alignment.
# Options are: 'compress pic', 'use whole pic' and 'calculate per layer'.
# We have been using 'compress pic', which  is the most accurate and
# fastest method.
# Therefore  'use whole pic' and 'calculate per layer' have not been
# extensively tested.
method = 'compress pic'


#######################Second level functions#######################
def align_single_pair(tiles, tile_file, contig_tuples, contig_ind,
                      micData, nr_peaks, nr_dim=2,
                      nr_slices=None):
    """Determine the ideal alignment between two neighbouring tiles

    Parameters:
    -----------
    tiles: list
        List of strings. List of references to the the tiles in the hdf5 file tile_file.
    tile_file: pointer
        hdf5 file object. The opened file containing the tiles to stitch.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    contig_ind: int
        The index of the current tile pair in
        contig_tuples. More precisely: the index of the
        tuple in contig_tuples containing the indexes of
        the tiles that should be aligned.
    micData: object
        MicroscopeData object. Should contain coordinates of
        the tile corners as taken from the microscope.
        These coordinates are used to dtermine the overlap
        between a tile pair.
    nr_peaks: int
        The n highest peaks from the PCM
        matrix that will be used to do crosscovariance
        with. A good number for 2D analysis is 8 peaks and
        good numbers for 3D with method ='compress pic' are
        6 or 9 peaks.
    nr_dim: int
        If 3, the code will assume three
        dimensional data for the tile, where z is the first
        dimension and y and x the second and third. For any
        other value 2-dimensional data is assumed.
        (default: 2)
    nr_slices: int
        Only applicable when running with 3D
        pictures and using 'compres pic' method. Determines
        the number of slices that are compressed together
        (compression in the z-direction). If None,
        all the slices are compressed together. Default:
        None

    Returns:
    --------

    best_trans: np.array
        1 by 2 or 3 array containing best found (z), y and x translation.
    best_cov: float
        The covariance of the overlap after translation of the overlap by best_trans.
    contig_ind: int
        The index of the used tile pair in
        contig_tuples. This is necessary to return when
        running on multiple processors/cores. More
        precisely: the index of the tuple in contig_tuples
        containing the indexes of the tiles that should be
        aligned.
    """

    # Rename tile indexes
    ind1 = min(contig_tuples[contig_ind])
    ind2 = max(contig_tuples[contig_ind])
    logger.info("Calculating pairwise alignment for indexes: {} and {}"
                .format(ind1, ind2))
    if nr_dim == 3:
        tile_1 = inout.load_tile_3D(tiles[ind1], tile_file)
        tile_2 = inout.load_tile_3D(tiles[ind2], tile_file)
    else:
        tile_1 = inout.load_tile(tiles[ind1], tile_file)
        tile_2 = inout.load_tile(tiles[ind2], tile_file)

    if (tile_1.size and tile_2.size):
        # Determine overlap
        overlap1, overlap2, plot_order = determine_overlap(ind1,
                                                           ind2, tile_1,
                                                           tile_2,
                                                           micData)
        logger.debug("Shape of overlap 1 and 2: {} {}"
                     .format(overlap1.shape, overlap2.shape))

        if nr_dim == 3 and method == 'compress pic':
            if nr_slices is None:
                nr_slices = overlap1.shape[0]
            best_trans, best_cov = align_single_compress_pic(overlap1,
                                                             overlap2,
                                                             nr_peaks,
                                                             nr_dim,
                                                             nr_slices,
                                                             plot_order)
        else:
            unr_pos_transistions = calculate_pos_shifts(
                overlap1, overlap2, nr_peaks, nr_dim)
            logger.debug("Possible transistions: {}".
                         format(unr_pos_transistions))
            # Do correlation over the found shifts:
            best_trans, best_cov = find_best_trans(unr_pos_transistions,
                                                   overlap1, overlap2,
                                                   plot_order)

        # Give some feedback
        logger.info("Best shift: {} covariance: {}".format(best_trans,
                                                           best_cov))
        logger.debug(
            "Best shift type: {} {} covariance type: {} contig_ind type: {}"
            .format(type(best_trans), type(best_trans[0]),
                    type(best_cov), type(contig_ind)))
    else:
        # One of the tiles is empty
        best_trans = np.zeros(nr_dim, dtype=int)
        best_cov = np.nan
        logger.info(
            "Best shift: {}. One of the neighbours is empty".format(
                best_trans))
    return np.array(best_trans,dtype='int16'), best_cov, contig_ind


def refine_single_pair(tiles, tile_file, contig_tuples, contig_ind,
                       micData, old_P, nr_peaks,
                       nr_dim=2, nr_slices=None):
    """Determine the ideal alignment between two neighbouring tiles.

    Uses an old alignment as starting point. Meant to use on smFISH
    signal data, where the old alignment is taken from the aligning of
    the nuclei staining.

    Parameters:
    -----------

    tiles: list
        List of strings. List of references to the the tiles in the hdf5 file tile_file.
    tile_file: pointer
        hdf5 file object. The opened file containing the tiles to stitch.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    contig_ind: int
        The index of the current tile pair in
        contig_tuples. More precisely: the index of the
        tuple in contig_tuples containing the indexes of
        the tiles that should be aligned.
    micData: object
        MicroscopeData object. Should contain coordinates of
        the tile corners as taken from the microscope.
        These coordinates are used to dtermine the overlap
        between a tile pair.
    old_P: dict
        An old pairwise alignment containing a key
        'P' containing a flattened list of 2D or 3D
        pairwise translations.
        And containing a key 'covs' containing the
        normalized cross covariance for each alignment.
    nr_peaks: int
        The n highest peaks from the PCM
        matrix that will be used to do crosscovariance
        with. A good number for 2D analysis is 8 peaks and
        good numbers for 3D with method ='compress pic' are
        6 or 9 peaks.
    nr_dim: int
        If 3, the code will assume three
        dimensional data for the tile, where z is the first
        dimension and y and x the second and third. For any
        other value 2-dimensional data is assumed.
        (default: 2)
    nr_slices: int
        Only applicable when running with 3D
        pictures and using 'compres pic' method. Determines
        the number of slices that are compressed together
        (compression in the z-direction). If None,
        all the slices are compressed together. Default:
        None

    Returns:
    --------

    best_trans: np.array
        1 by 2 or 3 array containing best found (z), y and x translation.
    best_cov: float
        The covariance of the overlap after translation of the overlap by best_trans.
    contig_ind: int
        The index of the used tile pair in
        contig_tuples. This is necessary to return when
        running on multiple processors/cores. More
        precisely: the index of the tuple in contig_tuples
        containing the indexes of the tiles that should be
        aligned.
    """

    # Rename the indexes of the 2 images to compare:
    ind1 = min(contig_tuples[contig_ind])
    ind2 = max(contig_tuples[contig_ind])
    logger.info("Calculating pairwise alignment for indexes: {} and {}"
                .format(ind1, ind2))
    if nr_dim == 3:
        tile_1 = inout.load_tile_3D(tiles[ind1], tile_file)
        tile_2 = inout.load_tile_3D(tiles[ind2], tile_file)
    else:
        tile_1 = inout.load_tile(tiles[ind1], tile_file)
        tile_2 = inout.load_tile(tiles[ind2], tile_file)


    if (tile_1.size and tile_2.size):
        # Determine overlap
        overlap1, overlap2, plot_order = determine_overlap(ind1, ind2,
                                                           tile_1,
                                                           tile_2,
                                                           micData)
        # Pick the old translation from te pairwise alignment
        # array
        trans = old_P[contig_ind * nr_dim: contig_ind * nr_dim + nr_dim]
        logger.debug("Cur old trans is: {}".format(trans))

        # Determine the overlap translated according to the old
        # pairwise translation
        overlap1_ref, overlap2_ref = ph.calc_translated_pics(
            trans,
            overlap1, overlap2,
            round_size=True)
        # Take just a bright part from this overlap.
        # TODO: code below may crash when possible transistions
        # are too big.
        # cut_out_coord  = ph.select_cut_out2(overlap1_ref, overlap2_ref)
        # overlap1_ref = overlap1_ref[cut_out_coord[0]:(cut_out_coord[0] + 150),
        #                    cut_out_coord[1]:(cut_out_coord[1] + 150)]
        # overlap2_ref = overlap2_ref[cut_out_coord[0]:(cut_out_coord[0] + 150),
        #                    cut_out_coord[1]:(cut_out_coord[1] + 150)]
        logger.debug("Shape of overlap 1 and 2: {} {}"
                     .format(overlap1_ref.shape, overlap2_ref.shape))

        if nr_dim == 3 and method == 'compress pic':
            if nr_slices is None:
                nr_slices = overlap1.shape()[0]
            best_trans, best_cov = align_single_compress_pic(
                overlap1_ref,
                overlap2_ref, nr_peaks, nr_dim, nr_slices,
                plot_order)
        else:
            # Calculate possible transistions
            unr_pos_transistions = calculate_pos_shifts(overlap1_ref,
                                                        overlap2_ref,
                                                        nr_peaks,
                                                        nr_dim)
            # Calculate best xcov of the found shifts:
            best_trans, best_cov = find_best_trans(unr_pos_transistions,
                                                   overlap1_ref,
                                                   overlap2_ref,
                                                   plot_order)
        # Give some feedback
        logger.info("Best shift: {} covariance: {}".
                    format(best_trans, best_cov))
    else:
        best_trans = np.zeros(nr_dim, dtype=int)
        best_cov = np.nan
        logger.info("Best shift: {}. One of the neighbours is empty"
                    .format(best_trans))

    return list(best_trans), best_cov, contig_ind


def align_single_compress_pic(overlap1, overlap2, nr_peaks, nr_dim,
                              nr_slices,
                              plot_order):
    """Perform the alignment when using the 3D method "compress pic"

    Parameters:
    -----------

    overlap1: np.array
        Image that overlaps with overlap2
    overlap2: np.array
        Image that overlaps with overlap1
    nr_peaks: int   
        The n highest peaks from the PCM
        matrix that will be used to do crosscovariance
        with. A good number for 2D analysis is 8 peaks and
        good numbers for 3D with method ='compress pic' are
        6 or 9 peaks.
    nr_dim: int
        If 3, the code will assume three
        dimensional data for the tile, where z is the first
        dimension and y and x the second and third. For any
        other value 2-dimensional data is assumed.
        (default: 2)
    nr_slices: int
        Only applicable when running with 3D
        pictures and using 'compres pic' method. Determines
        the number of slices that are compressed together
        (compression in the z-direction).
    plot_order: np.array
        Numpy array, filled with ones. The order in wich subplots should be made
        if we want to plot overlap1 and 2


    Returns:
    --------

    best_trans: np.array
        1 by 3 array containing best found z, y and x translation.
    best_cov: float
        The covariance of the overlap after translation of the overlap by best_trans.
    """
    # Unraveled possible transistions contains for each substack
    # a list which contains for each of the 3 dimensions
    logger.info("Calculating pairwise alignment using compress pic "
                + "method.")
    best_trans_list = []
    best_cov_list = []
    unr_pos_transistions = []
    counter = 0
    # Change number of peaks, because we are dealing with 3 dimensions
    nr_peaks = int(np.rint(nr_peaks / 3))
    logger.debug("nr_peaks data type: {}".format(type(nr_peaks)))
    while counter < overlap1.shape[0]:
        com_overlap1 = []
        com_overlap2 = []
        for i in range(overlap1.ndim):
            com_overlap1.append(
                np.amax(overlap1[counter:counter + nr_slices, :, :]
                        , axis=i))
            com_overlap2.append(
                np.amax(overlap2[counter:counter + nr_slices, :, :]
                        , axis=i))

            # plt.figure('compressed image')
            # plt.imshow(com_overlap1[-1])
            # plt.show()
        unr_pos_transistions.append(calculate_pos_shifts(
            com_overlap1, com_overlap2, nr_peaks, nr_dim))
        counter += nr_slices

        # Do correlation over the found shifts:
        best_compr_trans = np.zeros((len(unr_pos_transistions[-1]), 2),
                                    dtype=int)
        best_compr_cov = np.zeros(len(unr_pos_transistions[-1]))
        # Test all transistions
        for com_dim in range(len(unr_pos_transistions[-1])):
            best_compr_trans[com_dim], best_compr_cov[
                com_dim] = find_best_trans(
                unr_pos_transistions[-1][com_dim],
                com_overlap1[com_dim], com_overlap2[com_dim],
                plot_order)
        logger.debug('best_compr_trans: {}'
                     .format(best_compr_trans))
        # Find the best translation based on which translation has the highest covariance
        # Sort the covariance
        cov_order = np.argsort(best_compr_cov)
        logger.debug('cov_order: {}'
                     .format(cov_order))
        # The translation in each direction is hidden in the 3 flattened pictures.
        # The posible z trasistions are at best_compr_trans[1,0] and best_compr_cov[2,0]
        # The posible y trasistions are at best_compr_trans[0,0] and best_compr_cov[2,1]
        # The posible x trasistions are at best_compr_trans[0,1] and best_compr_cov[1,1]
        # Based get the row indexes for the best translation in the z, y and x direction:
        z_ind = cov_order[max(np.nonzero(cov_order == 1)[0][0],
                              np.nonzero(cov_order == 2)[0][0])]
        logger.debug('z_ind index in cov_order options: {} {}'
                     .format(np.nonzero(cov_order == 1)[0][0],
                             np.nonzero(cov_order == 2)[0][0]))
        y_ind = cov_order[max(np.nonzero(cov_order == 0)[0][0],
                              np.nonzero(cov_order == 2)[0][0])]
        logger.debug('y_ind index in cov_order options: {} {}'
                     .format(np.nonzero(cov_order == 0)[0][0],
                             np.nonzero(cov_order == 2)[0][0]))
        x_ind = cov_order[max(np.nonzero(cov_order == 0)[0][0],
                              np.nonzero(cov_order == 1)[0][0])]
        logger.debug('x_ind index in cov_order options: {} {}'
                     .format(np.nonzero(cov_order == 0)[0][0],
                             np.nonzero(cov_order == 1)[0][0]))
        logger.debug('z_ind, y_ind, x_ind: {} {} {}'
                     .format(z_ind, y_ind, x_ind))
        best_trans = np.zeros(3, dtype=int)
        # Z translation
        best_trans[0] = best_compr_trans[z_ind, 0]
        best_cov = best_compr_cov[z_ind]
        logger.debug('z compr cov: {} best cov overall: {}'
                     .format(best_compr_cov[z_ind], best_cov))
        # Y translation
        if y_ind == 0:
            best_trans[1] = best_compr_trans[y_ind, 0]
            best_cov = best_cov + best_compr_cov[y_ind]
        elif y_ind == 2:
            best_trans[1] = best_compr_trans[y_ind, 1]
            best_cov = best_cov + best_compr_cov[y_ind]
        else:
            logger.warning(
                'y_ind has an invalid value, gonna raise an error')
            raise IndexError(
                'y_ind has an invalid value, it should be 0 or 2, it is currently {}'
                .format(y_ind))
        logger.debug('y compr cov: {} best cov overall: {}'
                     .format(best_compr_cov[y_ind], best_cov))
        # X translation
        best_trans[2] = best_compr_trans[x_ind, 1]
        best_cov = (best_cov + best_compr_cov[x_ind]) / 3.0
        logger.debug('x compr cov: {} best cov overall: {}'
                     .format(best_compr_cov[x_ind], best_cov))
        best_trans_list.append(best_trans)
        best_cov_list.append(best_cov)
    # Pick the best
    logger.debug('best_trans_list: {}'.format(best_trans_list))
    best_trans = best_trans_list[np.argmax(best_cov_list)]
    best_cov = np.nanmax(best_cov_list)

    return best_trans, best_cov


def determine_overlap(ind1, ind2, tile_1, tile_2, micData):
    """Determine the overlap between two neighbouring tiles

    Parameters:
    -----------

    ind1: int
        Index (flattened) of tile 1
    ind2: int
        Index (flattened) of tile 2
    tile_1: np.array
        Image 1
    tile_2: np.array
        Image 2
    micData: object
        MicroscopeData object containing coordinates

    Returns:
    --------

    overlap1: np.array
        Overlapping part of tile_1
    overlap2: np.array
        Overlapping part of tile_2
    plot_order: np.array
        Numpy array of ones. The shape of this array is
        used for plotting the overlaps in well fitting
        subplots.
    """
    if abs(ind1 - ind2) == 1:
        # Overlap on left or right
        logger.info(("Calculating overlap: right of tile {0} and "
                      + "left of tile {1}").format(ind1, ind2))
        logger.debug(
            "Ind: {0} , tile nr: {1} , x-coord: {2}"
            .format(ind1, micData.tile_set.flat[:][ind1],
                    micData.x_coords[micData.tile_set.flat[:][ind1]]))
        logger.debug(
            "Ind: {0} , tile nr: {1} , x-coord: {2}"
            .format(ind2, micData.tile_set.flat[:][ind2],
                    micData.x_coords[micData.tile_set.flat[:][ind2]]))
        # Calculate overlap indexes
        overlap_ind_x = int(
            micData.x_coords[micData.tile_set.flat[:][ind1]]
            - micData.x_coords[micData.tile_set.flat[:][ind2]])
        overlap_ind_y = int(
            micData.y_coords[micData.tile_set.flat[:][ind1]]
            - micData.y_coords[micData.tile_set.flat[:][ind2]])
        logger.debug("Overlap index, x: {} y: {}".
                     format(overlap_ind_x, overlap_ind_y))
        overlap1, overlap2 = ph.get_overlapping_region(tile_1, tile_2,
                                                       overlap_ind_x,
                                                       overlap_ind_y,
                                                       'left')
        # For subplotting the overlaps later:
        plot_order = np.ones((1, 3))
    else:
        # Overlap on top or bottom
        logger.info(("Calculating overlap: bottom of tile {0} and "
                     + "top of tile {1}").format(ind1, ind2))
        logger.debug(
            "Ind: {0} , tile nr: {1} , y-coord: {2}"
            .format(ind1, micData.tile_set.flat[:][ind1],
                    micData.y_coords[micData.tile_set.flat[:][ind1]]))
        logger.debug(
            "Ind: {0} , tile nr: {1} , y-coord: {2}"
            .format(ind2, micData.tile_set.flat[:][ind2],
                    micData.y_coords[micData.tile_set.flat[:][ind2]]))
        # Calculate overlap indexes
        overlap_ind_y = int(
            micData.y_coords[micData.tile_set.flat[:][ind1]]
            - micData.y_coords[micData.tile_set.flat[:][ind2]])
        overlap_ind_x = int(
            micData.x_coords[micData.tile_set.flat[:][ind1]]
            - micData.x_coords[micData.tile_set.flat[:][ind2]])
        logger.debug("Overlap index, y: {} x: {}".
                     format(overlap_ind_y, overlap_ind_x))
        overlap1, overlap2 = ph.get_overlapping_region(tile_1, tile_2,
                                                       overlap_ind_x,
                                                       overlap_ind_y,
                                                       'top')
        # For subplotting the overlaps later:
        plot_order = np.ones((3, 1))

    return (overlap1, overlap2, plot_order)


def calculate_pos_shifts(overlap1, overlap2, nr_peaks, nr_dim):
    """Calulate possible shifts, given two overlapping images

    Parameters:
    -----------

    overlap1: np.array
        Image that overlaps with overlap2.
    overlap2: np.array
        Image that overlaps with overlap1.
    nr_peaks: int
        The number of peaks from the PCM that will be used to calculate shifts
    nr_dim: int
        If 3, the code will assume three
        dimensional data for the tile, where z is the first
        dimension and y and x the second and third. For any
        other value 2-dimensional data is assumed.

    Returns:
    --------
    unr_pos_transistions: np.array
        Numpy array or list (list only when
        method == 'compress pic' and nr_dim == 3 )
        Numpy array numpy arrays of int, with each
        of the inner arrays containing the (z), x
        and y translation, if nr_dim is not 3 only
        x and y translation are given.
        If method == 'compress pic' and nr_dim == 3
        a list of 3 lists is returned. In each list
        the best translations for each compressed
        picture are given as numpy arrays of length
        2.

    """
    # Calculate phase correlation matrix of the overlap
    if nr_dim == 3:
        logger.info("Calculating posible shifts, method: {}"
                    .format(method))
        if method == 'use whole pic':
            r1 = ph.calculate_PCM(overlap1, overlap2)
            # inout.plot_3D(r1)
        elif method == 'calculate per layer':
            r1 = ph.calculate_PCM_method2(overlap1, overlap2)
            # inout.plot_3D(r1)
        elif method == 'compress pic':
            logger.debug('length compressed overlap list: {}'.format(
                len(overlap1)))
            r1_list = []
            for i in range(len(overlap1)):
                r1 = ph.calculate_PCM(overlap1[i], overlap2[i])
                r1_list.append(r1)
                # inout.display_tiles(r1_list, np.ones((1,3)))
    else:
        logger.info("Calculating posible shifts, in 2D")
        r1 = ph.calculate_PCM(overlap1, overlap2)
    # Get the first nr_peaks peaks
    if method == 'compress pic' and nr_dim == 3:
        unr_pos_transistions = [[], [], []]
        collect_zeros = []
        for i in range(len(r1_list)):
            cur_trans = np.argsort(r1_list[i].flat[:])[-nr_peaks:]
            unr_cur_trans = np.array(
                np.unravel_index(cur_trans, r1_list[i].shape)).T
            logger.debug("unr_cur_trans: {}".format(unr_cur_trans))

            # Calculate correct transition from the found peaks (calculation as in Skimage)
            # noinspection PyTypeChecker
            midpoints = np.array([np.fix(axis_size / 2) for axis_size in
                                  overlap1[i].shape])
            logger.debug('Midpoints: {}'.format(midpoints))
            for trans in unr_cur_trans:
                logger.debug('trans bigger than midpoints: {}'.format(
                    trans > midpoints))
                trans[trans > midpoints] -= np.array(overlap1[i].shape)[
                    trans > midpoints]
                trans[trans < (-1 * midpoints)] \
                    += np.array(overlap1[i].shape)[
                    trans < (-1 * midpoints)]
                unr_pos_transistions[i].append(trans)
                # If trans is not all zeros, multiply transistion1 by -1
                # to get the transistions the both ways
                if trans.any():
                    unr_pos_transistions[i].append(-1 * trans)
            # Add zero translation if not already there
            collect_zeros += [pos for pos in unr_pos_transistions[i] if
                              not (np.any(pos))]
            logger.debug("Collect_zeros: {}".format(collect_zeros))
            if not (len(collect_zeros)):
                unr_pos_transistions[i].append(
                    np.zeros((len(unr_pos_transistions[i][0])),
                             dtype=int))
                logger.debug("Added zero trans to transistion")
                logger.debug(
                    "Possible transistion after appending zeros: {}"
                    .format(unr_pos_transistions[i]))
    else:
        pos_transistions1 = np.argsort(r1.flat[:])[-nr_peaks:]

        # Unravel the found shifts
        unr_pos_transistions1 = np.array(
            np.unravel_index(pos_transistions1, r1.shape)).T
        logger.debug("pos_trans1: {}".format(unr_pos_transistions1))

        if method == 'calculate per layer' and nr_dim == 3:
            # Calculate correct transition from the found peaks (calculation as in Skimage)
            logger.debug('Overlap shape for midpoints: {}'.format(
                overlap1.shape[-2:]))
            # noinspection PyTypeChecker
            midpoints = np.array([np.fix(axis_size / 2) for axis_size in
                                  overlap1.shape[-2:]])
            logger.debug('Midpoints: {}'.format(midpoints))
            for trans in unr_pos_transistions1:
                logger.debug('trans bigger than midpoints: {}'.format(
                    trans[-2:] > midpoints))
                trans[-2:][trans[-2:] > midpoints] -= \
                np.array(overlap1.shape)[-2:][trans[-2:] > midpoints]
                trans[-2:][trans[-2:] < (-1 * midpoints)] \
                    += np.array(overlap1.shape)[-2:][
                    trans[-2:] < (-1 * midpoints)]
        else:
            # Calculate correct transition from the found peaks
            # (calculation as in Skimage)
            midpoints = np.array(
                [np.fix(axis_size / 2) for axis_size in overlap1.shape])
            logger.debug('Midpoints: {}'.format(midpoints))
            for trans in unr_pos_transistions1:
                logger.debug('trans bigger than midpoints: {}'.format(
                    trans > midpoints))
                trans[trans > midpoints] -= np.array(overlap1.shape)[
                    trans > midpoints]
                trans[trans < (-1 * midpoints)] \
                    += np.array(overlap1.shape)[
                    trans < (-1 * midpoints)]
        # Multiply transistion1 by -1 to get the transistions the both
        # ways
        inv_transistion = -1 * unr_pos_transistions1
        # And remove all zero transistions from the inverted, because
        # these are duplicates
        inv_transistion = inv_transistion[
            ~np.all(inv_transistion == 0, axis=1)]
        unr_pos_transistions = np.vstack(
            (unr_pos_transistions1, inv_transistion))
        # Add zero translation if not already there
        collect_zeros = [pos for pos in unr_pos_transistions if
                         not (np.any(pos))]
        if not (len(collect_zeros)):
            unr_pos_transistions = np.append(unr_pos_transistions,
                                             np.zeros((1, len(
                                                 unr_pos_transistions[
                                                     0])), dtype=int),
                                             axis=0)
            logger.debug("Added zero trans to transistion")

    return unr_pos_transistions


def find_best_trans(pos_transistions, overlap1, overlap2, plot_order):
    """Find the best translation using the cross covariance.

    Shift overlap according to translations and test the cov of
    the translated overlaps.

    Parameters:
    -----------

    pos_transistions: np.array
        2D numpy array. Array containing y,x-pairs denoting the possible translations.
    overlap1: np.array
        Image
    overlap2: np.array
        Image that overlaps with overlap1.
    plot_order: np.array
        The shape of this array denotes
        the order in wich subplots should be made
        if we want to plot overlap1 and 2.

    Returns:
    --------

    best_trans: np.array
        1 by 2 or 3 array containing best found (z), y and x translation.
    best_cov: float
        The covariance of the overlap after translation of the overlap by best_trans.
    """
    # Init best_trans and best_cov
    # This makes sure that if all correlations are below threshold,
    # we get no translation.
    logger.info("Finding best translation.")

    best_trans = np.zeros(overlap1.ndim, dtype=int)  # np.array([0,0])
    best_cov = 0.0

    logger.debug('best_trans at start {}, type: {}'.format(best_trans,
                                                    type(best_trans)))

    for trans in pos_transistions:
        if method == 'calculate per layer' and overlap1.ndim == 3:
            shifted_a, shifted_b = ph.calc_translated_pics_3D(
                trans[-2:],
                overlap1, overlap2)
        elif method == 'use whole pic' and overlap1.ndim == 3:
            shifted_a, shifted_b = ph.calc_translated_pics_3D(trans[-2:],
                                                              overlap1,
                                                              overlap2)
        else:
            shifted_a, shifted_b = ph.calc_translated_pics(trans,
                                                           overlap1,
                                                           overlap2)
        # Calculate xcov:
        if shifted_a.ndim == 2:
            cov, monocolor = ph.xcov_nd(shifted_a, shifted_b)
            logger.debug('Found a 2D picture to compare')
        else:
            cov_list = []
            monocolor_list = []
            for i in range(shifted_a.shape[0]):
                cov, monocolor = ph.xcov_nd(shifted_a[i, :, :],
                                            shifted_b[i, :, :])
                cov_list.append(cov)
                monocolor_list.append(monocolor)
            cov_list = np.array(cov_list)
            monocolor_list = np.array(monocolor_list)
            cov = np.mean(cov_list)
            monocolor = monocolor_list.all()
            logger.debug("cov {} monocolor {}".format(cov, monocolor))
        # Check for monocolor images, they do not provide usefull cov
        if monocolor:
            logger.debug(
                "Monocolor image found, covariance for these images is zero")
            cov = 0
        if (cov > best_cov):
            best_cov = cov
            best_trans = trans
            # Give some feedback
            # inout.display_tiles([shifted_a, shifted_b], plot_order, fig_nr = 8, block = True)
    # Check the result
    thr = 0.5
    if best_cov < thr:
        best_trans = np.zeros(overlap1.ndim,
                              dtype=int)
    # Line to call display overlap if we want to plot the overlap with the
    # best covariance
    if overlap1.ndim == 2:
        ph.display_overlap(overlap1, overlap2, best_trans[-2:],
                           best_cov, plot_order)
    else:
        ph.display_overlap(overlap1[0, :, :], overlap2[0, :, :],
                           best_trans[-2:], best_cov, plot_order)
    logger.debug('best_trans at end {}, type: {}'.format(best_trans,
                                                         type(
                                                             best_trans)))
    return best_trans, best_cov


def plot_overlaps(alignment, tiles, contig_tuples, micData):
    """Plot the pairwise overlaps

    

    Parameters:
    -----------

    alignment: dict
        Dictionary containing key 'P' with a flattened list of translations.
    tiles: list
        List of strings. Each string points to a tile in the hdf5 file.
    contig_tuples: list
        List of tuples denoting which tiles are contingent to each other.
    micData: object
        MicroscopeData object containing coordinates.


    Notes:
    ------
    Should be tested and made to work in 3D ?_?

    """
    logger.info("Trying to plot overlaps, plot will only show when "
                + "display_overlap in pairwisehelper.py is True and "
                + "matplotlib is imported in inout.py.")
    for i in range(len(contig_tuples)):
        # Pick 2 images to compare:
        ind1 = min(contig_tuples[i])
        ind2 = max(contig_tuples[i])
        logger.debug("Current indexes: {}, {}".format(ind1, ind2))
        if (tiles[ind1].any() and tiles[ind2].any()):
            # Determine overlap
            overlap1, overlap2, plot_order = determine_overlap(
                ind1, ind2,
                tiles, micData, None)
            trans = alignment['P'][i * 2: i * 2 + 2]
            logger.debug("Cur trans to be checked: {}".format(trans))
            ph.display_overlap(overlap1, overlap2, trans, None,
                               plot_order)



def align_single_pair_npy(contig_tuple,filtered_files_list,micData, 
                          nr_peaks=8):
    
    """
    Determine the ideal alignment between two neighbouring tiles
    It is a modification of the align_single_pair
    function that will run in parallel using .npy image arrays.
    The functions runs only in 2D.
    
    Parameters:
    -----------

    contig_tuple: tuple
        Tuple containing two tile indexes denoting these
        tiles are contingent to each other.
    filtered_files_list: list
        List containing the paths to the files to porocess
    micData: object
        MicroscopeData object. Should contain coordinates of
        the tile corners as taken from the microscope.
        These coordinates are used to dtermine the overlap
        between a tile pair.
    nr_peaks: int
        The n highest peaks from the PCM
        matrix that will be used to do crosscovariance
        with. A good number for 2D analysis is 8 peaks and
        good numbers for 3D with method ='compress pic' are
        6 or 9 peaks.

    Returns:
    --------

    best_trans: np.array
        1 by 2 or 3 array containing best found (z), y and x translation.
    best_cov: float
        The covariance of the overlap after translation of the overlap by best_trans.
    contig_ind: int
        The index of the used tile pair in
        contig_tuples. This is necessary to return when
        running on multiple processors/cores. More
        precisely: the index of the tuple in contig_tuples
        containing the indexes of the tiles that should be
        aligned.
    """
    
    # This function run only for 2D so I can hard code the nr_dim
    nr_dim = 2
    
    # The contig_tuple contains the indexes derived from the tile_set

    ind1 = min(contig_tuple)
    ind2 = max(contig_tuple)
    
    logger.info("Calculating pairwise alignment for indexes: {} and {}"
                .format(ind1, ind2))
    
    # Load the images
    # Identify the tile positioned in the original image set
    tile_1_pos = micData.tile_set.flat[:][ind1]
    tile_2_pos = micData.tile_set.flat[:][ind2]
    
    
    
    tile_1_fpath = [fpath for fpath in filtered_files_list if 'pos_'+str(tile_1_pos)+'.' in fpath]
    tile_2_fpath = [fpath for fpath in filtered_files_list if 'pos_'+str(tile_2_pos)+'.' in fpath]
    if tile_1_fpath:
        tile_1 = np.load(tile_1_fpath[0])
        tile_1 =img_as_float(tile_1)
    else:
        tile_1 = np.array([],dtype=np.float64)

    if tile_2_fpath:
        tile_2 = np.load(tile_2_fpath[0])
        tile_1 =img_as_float(tile_1)
    else:
        tile_2 = np.array([],dtype=np.float64)
    
        
    if (tile_1.size and tile_2.size):
#         Determine overlap
        overlap1, overlap2, plot_order = determine_overlap(ind1,
                                                           ind2, tile_1,
                                                           tile_2,
                                                           micData)
    
        logger.debug("Shape of overlap 1 and 2: {} {}"
                     .format(overlap1.shape, overlap2.shape))
    
        unr_pos_transistions = calculate_pos_shifts(
                overlap1, overlap2, nr_peaks, nr_dim)

        # Do correlation over the found shifts:
        best_trans, best_cov = find_best_trans(unr_pos_transistions,
                                               overlap1, overlap2,
                                               plot_order)

         
        # Give some feedback
        logger.info("Best shift: {} covariance: {}".format(best_trans,
                                                           best_cov))
        logger.debug(
            "Best shift type: {} {} covariance type: {} contig_ind type: {}"
            .format(type(best_trans), type(best_trans[0]),
                    type(best_cov), type(contig_tuple)))
        
    else:
        # One of the tiles is empty
        best_trans = np.zeros(nr_dim, dtype=int)
        best_cov = np.nan
        
        logger.info(
            "Best shift: {}. One of the neighbours is empty".format(
                best_trans))
    
    return np.array(best_trans,dtype='int16'), best_cov, contig_tuple